import matplotlib
matplotlib.use('Agg')
matplotlib.rcParams['font.family']='serif'
matplotlib.rcParams['font.size']=18
matplotlib.rcParams['legend.fontsize']=15
matplotlib.rcParams['xtick.major.size'] = 10.0
matplotlib.rcParams['xtick.minor.size'] = 5.0
matplotlib.rcParams['ytick.major.size'] = 10.0
matplotlib.rcParams['ytick.minor.size'] = 5.0

import pylab
import scipy.stats
import numpy as np
import os
import sys
import collections

from .kde import KDE
from .plotter import Plotter

import scipy.optimize
try:
    from cosmosis import output as output_module
except ImportError:
    print("Running without cosmosis: no pretty section names or running on ini files")

class ChainPlotter(Plotter):

    @classmethod
    def from_outputs(cls, options, **kw):
        burn = kw.pop("burn", 0)
        thin = kw.pop("thin", 1)
        
        column_names, chains, metadata, comments, final_metadata = output_module.input_from_options(options)

        if burn==0:
            pass
        elif burn<1:
            for i,chain in enumerate(chains):    
                print("Burning fraction %f of chain %d, which is %d samples" %(burn,i,int(burn*len(chain[:,0]))))
            chains = [chain[int(burn*len(chain[:,0])):, :] for chain in chains]
        else:
            burn = int(burn)
            chains = [chain[burn:,:] for chain in chains]

    #In this case all the chains are assumed to be from a single
    #run.  So we should concatenate them all for a single 
        chains = np.vstack(chains).T
        chains = dict(list(zip(column_names,chains)))
        chain_data = {"Chains":chains}
        return cls(chain_data, **kw)


    @classmethod
    def from_chain_files(cls, filenames, burn=0, thin=1, **kw):
        nameset = [open(filename).readline().strip().strip('#').replace("/","--").split() for filename in filenames]
        dataset = [(np.loadtxt(filename).T) for filename in filenames]
        if burn==0:
            pass
        elif burn<1:

            for name,data in zip(filenames,dataset):    
                print("Burning fraction %f of chain %s, which is %d samples" %(burn,name,int(burn*len(data[0]))))
            dataset = [data[:,int(burn*len(data[0])):] for data in dataset]
        else:
            burn = int(burn)
            dataset = [data[:,burn:] for data in dataset]

        if thin!=1:
            dataset = [data[:,::thin] for data in dataset]

        chain_data = collections.OrderedDict()
        for filename,names,data in zip(filenames,nameset,dataset):
            chain_datum = dict(list(zip(names,data)))
            chain_data[filename] = chain_datum    
        return cls(chain_data, **kw)

    def _plot_1d(self, name, xmin_input=None, xmax_input=None, n=100, factor=2.0):
        for i,(filename, x) in enumerate(self.cols_for_name(name).items()):
            if xmin_input is None:
                xmin = x.min()
            else:
                xmin=xmin_input[i]
            if xmax_input is None:
                xmax = x.max()
            else:
                xmax=xmax_input[i]

            kde = KDE(x, factor=factor)
            x_axis, like = kde.grid_evaluate(n, (xmin,xmax) )
            #need to save plot_data called "like" here
            pylab.plot(x_axis, like, '-', color=self.colors[i], **self.plot_keywords_1d())
        pylab.xlabel("$"+self._display_names[name]+"$")
        if self.blind: self.blind_axes()
    
    @staticmethod
    def _find_contours_corrected(like, x, y, n, xmin, xmax, ymin, ymax, contour1, contour2):
        N = len(x)
        x_axis = np.linspace(xmin, xmax, n+1)
        y_axis = np.linspace(ymin, ymax, n+1)
        histogram, _, _ = np.histogram2d(x, y, bins=[x_axis, y_axis])


        def objective(limit, target):
            w = np.where(like>limit)
            count = histogram[w]
            return count.sum() - target
        target1 = N*(1-contour1)
        target2 = N*(1-contour2)
        level1 = scipy.optimize.bisect(objective, like.min(), like.max(), args=(target1,))
        level2 = scipy.optimize.bisect(objective, like.min(), like.max(), args=(target2,))
        return level1, level2, like.sum()


    @staticmethod
    def _find_contours(like, contour1, contour2):
        total_mass = like.sum()
        like_sorted = np.sort(like.flatten())
        like_cumsum = like_sorted.cumsum()
        height1 = np.interp(contour1/total_mass,like_cumsum,like_sorted)
        height2 = np.interp(contour2/total_mass,like_cumsum,like_sorted)
        return height1, height2, total_mass


    def _plot_2d(self, name1, name2, xmin_input=None, xmax_input=None, ymin_input=None, ymax_input=None, n=100, factor=2.0):
        cols1 = self.cols_for_name(name1)
        cols2 = self.cols_for_name(name2)
        fill = self.options.get('fill',True)
        for i,(filename,x) in enumerate(cols1.items()):
            if filename not in cols2: continue
            y = cols2[filename]
            if xmin_input is None:
                xmin = x.min()
            else:
                xmin=xmin_input[i]
            if xmax_input is None:
                xmax = x.max()
            else:
                xmax=xmax_input[i]

            if ymin_input is None:
                ymin = y.min()
            else:
                ymin=ymin_input[i]
            if ymax_input is None:
                ymax = y.max()
            else:
                ymax=ymax_input[i]

            kde = KDE([x,y], factor=factor)
            (x_axis,y_axis), like = kde.grid_evaluate(n, [(xmin,xmax),(ymin,ymax)] )
            like/=like.sum()
            contour1=1-0.68
            contour2=1-0.95
            level1, level2, total_mass = self._find_contours_corrected(like, x, y, n, xmin, xmax, ymin, ymax, contour1, contour2)
            level0 = 1.1
            levels = [level2, level1, level0]
            color=self.colors[i]
            if fill:
                pylab.contourf(x_axis, y_axis, like.T, [level2,level0], colors=[color], alpha=0.25)
                pylab.contourf(x_axis, y_axis, like.T, [level1,level0], colors=[color], alpha=0.25)
            else:
                pylab.contour(x_axis, y_axis, like.T, [level2,level1], colors=color)
        pylab.xlabel("$"+self._display_names[name1]+"$")
        pylab.ylabel("$"+self._display_names[name2]+"$")
        if "limits" in self.options and name1    in self.options["limits"]:
            pylab.xlim(*self.options["limits"][name1])
        if "limits" in self.options and name1    in self.options["limits"]:
            pylab.ylim(*self.options["limits"][name2])
        if "truth" in self.options and name1 in self.options["truth"] and name2 in self.options["truth"]:
            xc = self.options["truth"][name1]
            yc = self.options["truth"][name2]
            pylab.plot([xc],[yc],'k*', markersize=10)
        if self.blind: self.blind_axes()
        # self.command('pylab.xlabel("${0}$")',self._display_names[name2])




    def plot_keywords_1d(self):
        return dict(lw=5)

    def w0_wa_plot(self):
        print("Doing W0 - WA plot")
        w_name  = 'cosmological_parameters--w'
        wa_name = 'cosmological_parameters--wa'
        self._plot_2d(w_name, wa_name, fill=self.options.get('fill',True), factor=1.5)
        w_min,w_max = self.parameter_range(w_name)
        wa_min,wa_max = self.parameter_range(wa_name)
        w_axis = np.linspace(w_min,w_max,2)
        wa_axis = np.linspace(wa_min,wa_max,2)
        pylab.plot(w_axis,[0.0, 0.0],'--',color='gray',dashes=(20,20))
        pylab.plot([-1.0, -1.0], wa_axis,'--',color='gray',dashes=(20,20))
        pylab.plot([-1.0],[0.0],'k*',markersize=10)

        output_name = "%s/%sW0_WA.%s" % (self.root_dir,self.prefix,self.filetype,)
        pylab.savefig(output_name)
        pylab.close()



from cosmosis.runtime.config import Inifile
from cosmosis.samplers.sampler import Sampler
import cosmosis.samplers.minuit.minuit_sampler
from cosmosis.runtime.pipeline import LikelihoodPipeline
from cosmosis.output.in_memory_output import InMemoryOutput
import tempfile
import os
import sys
import pytest
import numpy as np

minuit_compiled = os.path.exists(cosmosis.samplers.minuit.minuit_sampler.libname)

def run(sampler, check_prior, check_extra=True, **options):

    sampler_class = Sampler.registry[sampler]

    values = tempfile.NamedTemporaryFile('w')
    values.write(
        "[parameters]\n"
        "p1=-3.0  0.0  3.0\n"
        "p2=-3.0  0.0  3.0\n")
    values.flush()

    override = {
        ('runtime', 'root'): os.path.split(os.path.abspath(__file__))[0],
        ("pipeline", "debug"): "F",
        ("pipeline", "quiet"): "T",
        ("pipeline", "modules"): "test1",
        ("pipeline", "extra_output"): "parameters/p3",
        ("pipeline", "values"): values.name,
        ("test1", "file"): "test_module.py",
    }

    for k,v in options.items():
        override[(sampler,k)] = str(v)


    ini = Inifile(None, override=override)

    # Make the pipeline itself
    pipeline = LikelihoodPipeline(ini)

    output = InMemoryOutput()
    sampler = sampler_class(ini, pipeline, output)
    sampler.config()


    while not sampler.is_converged():
        sampler.execute()

    if check_prior:
        pr = np.array(output['prior'])
        # a few samples might be outside the bounds and have zero prior
        assert np.all((pr==-3.58351893845611)|(pr==-np.inf))
        # but not all of them
        assert not np.all(pr==-np.inf)

    if check_extra:
        p1 = output['parameters--p1']
        p2 = output['parameters--p2']
        p3 = output['PARAMETERS--P3']
        assert np.all((p1+p2==p3)|(np.isnan(p3)))

    return output


def test_apriori():
    run('apriori', True, nsample=100)

def test_dynesty():
    # dynesty does not support extra params
    run('dynesty', False, check_extra=False, nlive=50, sample='unif')

def test_emcee():
    run('emcee', True, walkers=8, samples=100)

def test_fisher():
    run('fisher', False, check_extra=False)

def test_grid():
    run('grid', True, nsample_dimension=10)

def test_gridmax():
    run('gridmax', True, max_iterations=1000)

# def test_kombine():
#     run('kombine')

def test_maxlike():
    run('maxlike', True)

def test_metropolis():
    run('metropolis', True, samples=20)

@pytest.mark.skipif(not minuit_compiled,reason="requires Minuit2")
def test_minuit():
    run('minuit', True)

def test_multinest():
    run('multinest', True, max_iterations=10000, live_points=50, feedback=False)

def test_pmaxlike():
    run('pmaxlike', True)

def test_pmc():
    old_settings = np.seterr(invalid='ignore', divide='ignore')
    run('pmc', True, iterations=10)
    np.seterr(**old_settings)  

def test_zeus():
    run('zeus', True, walkers=10, samples=100, nsteps=50)
    run('zeus', True, walkers=10, samples=100, nsteps=50, verbose=True)
    run('zeus', True, walkers=10, samples=100, nsteps=50, tune=False)
    run('zeus', True, walkers=10, samples=100, nsteps=50, tolerance=0.1)
    run('zeus', True, walkers=10, samples=100, nsteps=50, patience=5000)
    run('zeus', True, walkers=10, samples=100, nsteps=50, maxiter=5000)
    run('zeus', True, walkers=10, samples=100, nsteps=50, moves="differential:2.0  global")

def test_polychord():
    with tempfile.TemporaryDirectory() as base_dir:
        run('polychord', True, live_points=20, feedback=0, base_dir=base_dir, polychord_outfile_root='pc')

def test_pymc():
    if sys.version_info.major==2:
        run('pymc', False)
    else:
        pytest.skip("Skipping pymc test on python 3")

def test_snake():
        run('snake', True)

def test_star():
        run('star', False)

def test_test():
    run('test', False)

if __name__ == '__main__':
    import sys
    locals()[sys.argv[1]]()


from .parameter import Parameter, register_new_parameter
from .config import CosmosisConfigurationError, Inifile
from .module import Module, SetupError, FunctionModule, ClassModule
from .prior import Prior
from .pipeline import LikelihoodPipeline, MissingLikelihoodError
from .mpi_pool import MPIPool
from .utils import stdout_redirected



import abc
from . import elements
from . import plots
from . import statistics
import numpy as np
from cosmosis import output as output_module
from ..runtime.config import Inifile
import imp
import os
from future.utils import with_metaclass
postprocessor_registry = {}

from io import StringIO


class PostProcessMetaclass(abc.ABCMeta):
    def __init__(cls, name, b, d):
        abc.ABCMeta.__init__(cls, name, b, d)
        sampler = d.get("sampler")
        if d is None: return
        postprocessor_registry[sampler] = cls

class PostProcessor(with_metaclass(PostProcessMetaclass, object)):
    sampler=None
    cosmosis_standard_output=True
    def __init__(self, ini, label, index, **options):
        super(PostProcessor,self).__init__()
        self.options=options
        self.sampler_options={}
        self.steps = []
        self.index = index
        self.label = label
        self.derive_file = options.get("derive", "")
        self.load(ini)
        elements = [el for el in self.elements if (not issubclass(el, plots.Plots) or (not options.get("no_plots")))]
        self.steps = [e(self, **options) for e in elements]
        self.outputs = {} #outputs can be anything, but for now matplotlib figures and open text files

    def load_extra_steps(self, filename):
        extra = elements.PostProcessorElement.instances_from_file(filename, self, **self.options)
        for e in extra:
            print("Adding post-processor step: %s" % (e.__class__.__name__))
        self.steps.extend(extra)

    def add_rerun_bestfit_step(self, dirname):
        from .reruns import BestFitRerunner
        rerunner = BestFitRerunner(dirname, self, **self.options)
        self.steps.append(rerunner)



    def approximate_scale_ceiling(self, c):
        r = np.mean([d[:,c].std() for d in self.data])
        scale = (10**np.ceil(np.log10(abs(r))))
        return scale

    def additive_blind_column(self, c, value):
        for d in self.data:
            d[:,c] += value

    def multiplicative_blind_column(self, c, value):
        for d in self.data:
            d[:,c] *= (1+value)

    def derive_extra_column(self, function):
        new_data = []
        for d in self.data:
            chain = SingleChainData(d,self.colnames)
            col, code = function(chain)
            if col is None:
                break
            #insert a new column into the chain, second from the end
            d = np.insert(d, -2, col, axis=1)
            #save the new chain
            new_data.append(d)
        if code is None:
            return

        self.colnames.insert(-2, code)
        self.data = new_data



    def derive_extra_columns(self):
        if not self.derive_file: return
        name = os.path.splitext(os.path.split(self.derive_file)[1])[0]
        module = imp.load_source(name, self.derive_file)
        functions = [getattr(module,f) for f in dir(module) if f.startswith('derive_')]
        print("Deriving new columns from these functions in {}:".format(self.derive_file))
        for f in functions:
            self.derive_extra_column(f)
    def load_tuple(self, inputs):
        self.colnames, self.data, self.metadata, self.comments, self.final_metadata = inputs
        self.name = "Data"
        for chain in self.metadata:
            for key,val in list(chain.items()):
                self.sampler_options[key] = val

    def load_dict(self, inputs):
        output_options=inputs["output"]
        filename = output_options['filename']
        self.name = filename
        sampler = inputs['sampler']
        self.colnames, self.data, self.metadata, self.comments, self.final_metadata = inputs['data']
        for chain in self.metadata:
            for key,val in list(chain.items()):
                self.sampler_options[key] = val

    def load_ini(self, inputs):
        output_options = dict(inputs.items('output'))
        filename = output_options['filename']
        self.name = filename
        sampler = inputs.get("runtime", "sampler")
        self.colnames, self.data, self.metadata, self.comments, self.final_metadata = \
            output_module.input_from_options(output_options)
        for chain in self.metadata:
            for key,val in list(chain.items()):
                self.sampler_options[key] = val


    def sampler_option(self, key, default=None):
        return self.sampler_options.get(key, default)

    def load_chain(self, ini):
        if isinstance(ini, tuple):
            self.load_tuple(ini)
        elif isinstance(ini, dict):
            self.load_dict(ini)
        else:
            self.load_ini(ini)

        #derive any additional parameters
        self.derive_extra_columns()

        #set the column names
        self.colnames = [c.lower() for c in self.colnames]
        self.data_stacked = np.concatenate(self.data).T

    def load(self, ini):
        if self.cosmosis_standard_output:
            self.load_chain(ini)

    def extract_ini(self, tag):
        in_ = False
        lines = []
        for line in self.comments[0]:
            line = line.strip()
            if line == 'START_OF_{}_INI'.format(tag):
                in_ = True
            elif line == 'END_OF_{}_INI'.format(tag):
                break
            elif in_:
                lines.append(line)

        s = StringIO(u"\n".join(lines))
        ini = Inifile(None)
        ini.read_file(s)
        return ini



        

    def __len__(self):
        return self.data_stacked.shape[1]

    def get_row(self, index):
        return self.data_stacked[:,index]

    def has_col(self, name):
        return name in self.colnames

    def get_col(self, index_or_name, stacked=True):
        """Get the named or numbered column."""
        if isinstance(index_or_name, int):
            index = index_or_name
        else:
            name = index_or_name
            index = self.colnames.index(name)
        cols = [d[:,index] for d in self.data]
        if stacked:
            return np.concatenate(cols)
        else:
            return cols

    def run(self):
        files = []
        for e in self.steps:
            try:
                files += e.run()
            except KeyboardInterrupt:
                raise
            except:
                import traceback
                if self.options.get("pdb", False):
                    print()
                    import pdb
                    pdb.post_mortem()
                else:                    
                    print("Failed in one of the postprocessing steps: ", e)
                    print("Here is the error stack:")
                    print(traceback.format_exc())
        return files

    def finalize(self):
        print("Finalizing:")
        for e in self.steps:
            e.finalize()
        for f in list(self.outputs.values()):
            print("Output: ", f.filename)
            f.finalize()

    def apply_tweaks(self, tweaks):
        if tweaks.filename==plots.Tweaks.filename:
            print(tweaks.filename)
            print("Please fill in the 'filename' attribute of your tweaks")
            print("Put the base name (without the directory, prefix, or suffix)")
            print("of the filename you want to tweak.")
            print("You can use also use a list for more than one plot,")
            print("or put '%s' to apply to all plots."%plots.Tweaks._all_filenames)
            return
        elif tweaks.filename==tweaks._all_filenames:
            filenames = [o.name for o in list(self.outputs.values()) if isinstance(o, plots.PostprocessPlot)]
        elif isinstance(tweaks.filename, list):
                filenames = tweaks.filename
        else:
            filenames = [tweaks.filename]
        for output in list(self.outputs.values()):
            if output.name in filenames:
                output.tweak(tweaks)



class SingleChainData(object):
    """
    This helper object is to make it easier for users to write functions
    that derive new parameters.
    """
    def __init__(self, data, colnames):
        self.data = data
        self.colnames = colnames

    def __getitem__(self, index_or_name):
        if isinstance(index_or_name, int):
            index = index_or_name
        else:
            name = index_or_name
            index = self.colnames.index(name)
        return self.data[:,index]



import os
import sys
import argparse
import subprocess

parser = argparse.ArgumentParser(description="print commands that set up the cosmosis env")
cosmosis_src_dir = os.path.split(__file__)[0]
parser.add_argument("--source", default=cosmosis_src_dir)
parser.add_argument("--no-omp", action='store_false', dest='omp', help='Switch off OpenMP')
parser.add_argument("--omp", action='store_true', dest='omp', help='Switch on OpenMP (default)')
parser.add_argument("--debug", action='store_true', dest='debug', help='Switch on debug mode')
parser.add_argument("--no-debug", action='store_false', dest='debug', help='Switch on debug mode')
parser.add_argument("--no-conda", action='store_false', dest='conda', help='Switch off conda flags, even if conda env is found')
parser.add_argument("--brew", action='store_true', help='Print commands for homebrew with clang')
parser.add_argument("--brew-gcc", action='store_true', help='Print commands for homebrew with gcc')

def homebrew_gfortran_libs():
    s = subprocess.run('gfortran -print-search-dirs', shell=True, capture_output=True)
    if s.returncode:
        return ""
    s.stdout.decode().split("\n")

    for line in lines:
        if line.startswith("libraries:"):
            break
    else:
        return ""
    try:
        libdir = line.split("=")[1].split(":")[-1]
    except:
        return ""

    return f"-L {libdir}"

def homebrew_gcc_commands():
    s = subprocess.run('brew list --versions gcc', shell=True, capture_output=True)
    version = s.stdout.decode().split()[1].split('.')[0]
    return [
        f"export CC=gcc-{version}",
        f"export CXX=g++-{version}",
        f"export FC=gfortran-{version}",
        "export MPIFC=mpif90",
        "export COSMOSIS_ALT_COMPILERS=1",
    ]



def generate_commands(cosmosis_src_dir, debug=False, omp=True, brew=False, brew_gcc=False, conda=True):
    conda = conda and ("CONDA_PREFIX" in os.environ)

    commands = [
        f"export COSMOSIS_SRC_DIR=\"{cosmosis_src_dir}\"",
        "export LIBRARY_PATH=$LIBRARY_PATH:$COSMOSIS_SRC_DIR/datablock",
        "export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$COSMOSIS_SRC_DIR/datablock",
        "export DYLD_LIBRARY_PATH=$DYLD_LIBRARY_PATH:$COSMOSIS_SRC_DIR/datablock",
        "export COSMOSIS_ALT_COMPILERS=1",
    ]

    if brew:
        commands += [
            "export GSL_LIB=/usr/local/lib",
            "export GSL_INC=/usr/local/include",
            "export FFTW_LIBRARY=/usr/local/lib",
            "export FFTW_INCLUDE_DIR=/usr/local/include",
            "export LAPACK_LINK='-L /usr/local/opt/openblas/lib/ -l lapack'",
            "export LAPACK_LIB=/usr/local/opt/openblas/lib/",
            "export CFITSIO_LIB=/usr/local/lib",
            "export CFITSIO_INC=/usr/local/include",
        ]

        if brew_gcc:
            commands += homebrew_gcc_commands()
        else:
            commands += [
                f"export CC=clang",
                f"export CXX=clang++",
                f"export FC=gfortran",
                "export MPIFC=mpif90",
                "export COSMOSIS_ALT_COMPILERS=1",
            ]

    elif conda:
        commands += [
            'export GSL_LIB=$CONDA_PREFIX/lib',
            'export GSL_INC=$CONDA_PREFIX/include',
            'export FFTW_LIBRARY=$CONDA_PREFIX/lib',
            'export FFTW_INCLUDE_DIR=$CONDA_PREFIX/include',
            'export LAPACK_LINK="-L$CONDA_PREFIX/lib -llapack"',
            'export LAPACK_LIB="$CONDA_PREFIX/lib"',
            'export CFITSIO_LIB=$CONDA_PREFIX/lib',
            'export CFITSIO_INC=$CONDA_PREFIX/include',
        ]


    if omp:
        commands.append("export COSMOSIS_OMP=1")
        
    if debug:
        commands.append("COSMOSIS_DEBUG=1")

    return commands

if __name__ == '__main__':
    args = parser.parse_args()
    cmds = generate_commands(args.source, debug=args.debug, omp=args.omp, conda=args.conda, brew=args.brew or args.brew_gcc, brew_gcc=args.brew_gcc)
    for command in cmds:
        print(command)




from .utils import std_weight
import numpy as np





def smooth_density_estimate_1d(samples, xmin, xmax, weights=None, N=1024, smoothing=1, fix_boundary=True):
    """
    Generate a smooth estimate of a 1D PDF from some samples using Kernel Density Estimation,
    correcting at the boundaries.
    
    The boundary fix algorithm comes from Jones & Foster 1996, by way of the getdist code,
    which this implementation follows.  Like that code, it uses convolutions to do the main calculations.
    Without the fix, parameters whose posterior has a hard boundary that cuts off the PDF
    where it is still large will have a spurious decrease in the smoothed PDF.
    
    If the boundary fix is applied, the upper and lower limit are assumed to be
    hard cut-offs in the PDF, not just the range where you happen to want the plots.
    
    Parameters
    ----------
    samples: 1D array
        samples from the PDF
    xmin: float
        lower bound, used for both the choice of output range and the boundary fix (if active)
    xmax: float
        upper bound, ditto
    weights: array
        optiona, default None. Weights per sample
    N: int
        optional, default 1024.  Number of sample points to return.
        Much faster if a power of 2
    smoothing: float/int
        optional, default 1.  Widening of the kernel, to smooth more
    fix_boundary: bool
        optional, default True.  Apply the fix to the boundaries
    """
    from scipy.signal import fftconvolve

    # Compute the things we need for the bandwidth choice:
    if weights is None:
        neff = len(samples)
        stdev = samples.std()
    else:
        neff = weights.sum() ** 2 / (weights**2).sum()
        stdev = std_weight(samples, weights)

    # This is a standard factor for 1D KDEs
    scott_factor = neff**(-0.2)

    # We optionally allow additional smoothing specified by the user
    width = stdev * scott_factor * smoothing

    # First we get the raw histogram that forms the basis of
    # the rest of this.
    edges =  np.linspace(xmin, xmax, N+1)
    x = 0.5 * (edges[1:] + edges[:-1])
    P_histogram, _ = np.histogram(samples, bins=edges, weights=weights)
    w = (xmax - xmin) / N

    # smoothing scale in units of the bin width
    s = width / w

    # Make the Gaussian kernel with which we are convolving.
    # We go out to 3 sigma
    window_width = int(3 * s)
    window_x = np.arange(-window_width, window_width+1)
    kernel = np.exp(-0.5 * (window_x/s)**2)
    
    # Generate the smoothed version.  If we do not need
    # the boundary smoothing then this is our final output
    P_smooth = fftconvolve(P_histogram, kernel, 'same')
    
    # If so, normalize and return it here
    if not fix_boundary:
        P_smooth /= P_smooth.sum() * (x[1] - x[0])        
        return x, P_smooth

    # Generate the mask, a top-hat which cuts off where the
    # boundaries are
    full_width = N + 2 * window_width
    mask = np.ones(full_width)
    mask[:window_width] = 0
    mask[window_width] = 0.5
    mask[-window_width:] = 0
    mask[-(window_width+1)] = 0.5

    # Apply the correction, which uses
    # modified versions of the convolution kernel to
    # get better estimates of the edge region
    
    a0 = fftconvolve(mask, kernel, 'valid')
    
    # Avoid a divide-by-zero
    ix = np.nonzero(a0 * P_smooth)
    a0 = a0[ix]
    
    # Main correction calculation
    P_norm = P_smooth[ix] / a0
    xK = window_x * kernel
    x2K = xK * window_x
    a1 = fftconvolve(mask, xK, mode='valid')[ix]
    a2 = fftconvolve(mask, x2K, mode='valid')[ix]
    xP = fftconvolve(P_histogram, xK, mode='same')[ix]

    # Apply the correction, which is done inside the exponential
    # to keep everything positive.
    scaling = (P_smooth[ix] * a2 - xP * a1) / (a0 * a2 - a1 ** 2)
    P_final = P_smooth.copy()
    P_final[ix] = P_norm * np.exp(np.minimum(scaling / P_norm, 4) - 1)

    # Normalize and return
    P_final /= P_final.sum() * (x[1] - x[0])
    return x, P_final

def smooth_density_estimate_2d(x, y, xmin, xmax, ymin, ymax, weights=None, N=256, smoothing=1, fix_boundary=True):
    """
    Generate a smooth estimate of a 2D PDF from some samples using Kernel Density Estimation,
    
    The boundary fix algorithm comes from Jones & Foster 1996, by way of the getdist code,
    which this implementation follows.  Like that code, it uses convolutions to do the main calculations.
    Without the fix, parameters whose posterior has a hard boundary that cuts off the PDF
    where it is still large will have a spurious decrease in the smoothed PDF.
    
    If the boundary fix is applied, the upper and lower limit are assumed to be
    hard cut-offs in the PDF, not just the range where you happen to want the plots.
    
    Parameters
    ----------
    x: 1D array
        samples from one parameter of the PDF
    y: 1D array, corresponding samples from a second parameter
    xmin: float
        lower bound for x used for both the choice of output range and the boundary fix (if active)
    xmax: float
        upper bound, ditto
    ymin: float
        lower bound for y used for both the choice of output range and the boundary fix (if active)
    ymax: float
        upper bound, ditto
    weights: array
        optiona, default None. Weights per sample
    N: int
        optional, default 256.  Min number of sample points to return  per dimension.
        Much faster if a power of 2
    smoothing: float/int
        optional, default 1.  Widening of the kernel, to smooth more
    fix_boundary: bool
        optional, default True.  Apply the fix to the boundaries
    """
    from scipy.signal import fftconvolve

    # Compute the things we need for the bandwidth choice:
    if weights is None:
        neff = len(x)
        covmat = np.cov([x, y])
    else:
        neff = weights.sum() ** 2 / (weights**2).sum()
        covmat = np.cov([x, y], aweights=weights)

    # make the correlation matrix from the covariance
    corr = covmat.copy()
    for i in range(2):
        si = corr[i, i]**0.5
        corr[i, :] /= si
        corr[:, i] /= si
    rho = corr[0, 1]
    
    if rho > 0.6:
        N = max(512, N)
        
    # This is a standard factor for 1D KDEs
    scott_factor = neff**(-1/6.)

    # First we get the raw histogram that forms the basis of
    # the rest of this.
    xedges = np.linspace(xmin, xmax, N+1)
    yedges = np.linspace(ymin, ymax, N+1)
    P_histogram, _, _ = np.histogram2d(x, y, bins=[xedges, yedges], weights=weights)
    xmid = 0.5 * (xedges[1:] + xedges[:-1])
    ymid = 0.5 * (yedges[1:] + yedges[:-1])
    
    dx = (xmax - xmin) / N
    dy = (ymax - ymin) / N

    # smoothing scale in units of the bin width
    width_x = np.sqrt(covmat[0, 0]) * scott_factor * smoothing / dx
    width_y = np.sqrt(covmat[1, 1]) * scott_factor * smoothing / dy
    
    # get the smoothing kernel.  Much nicer if you don't have to support py2!
    kernel_C = np.array([[width_x**2, width_x * width_y * rho], [width_x * width_y * rho, width_y**2]])
    kernel_Cinv = np.linalg.inv(kernel_C)

    # Make the Gaussian kernel with which we are convolving.
    # We go out to 3 sigma
    window_width = int(3 * max(width_x, width_y))
    
    window_xy = np.mgrid[-window_width:window_width + 1, -window_width:window_width + 1]
    r2 = np.einsum('kij,kl,lij->ij', window_xy, kernel_Cinv, window_xy)
    kernel = np.exp(-0.5*r2)
    kernel /= kernel.sum()
    
    # Generate the smoothed version.  If we do not need
    # the boundary smoothing then this is our final output
    P_smooth = fftconvolve(P_histogram, kernel, 'same')
    
    # If so, normalize and return it here
    if not fix_boundary:
        P_smooth /= P_smooth.sum() * dx * dy
        return xmid, ymid, P_smooth

    # Generate the mask, a top-hat which cuts off where the
    # boundaries are
    full_width = N + 2 * window_width
    mask1d = make_1d_mask(N, window_width)
    mask = np.outer(mask1d, mask1d)
    
    # Apply the correction, which uses
    # modified versions of the convolution kernel to
    # get better estimates of the edge PDF
    a00 = fftconvolve(mask, kernel, 'valid')
    
    # Avoid a divide-by-zero
    nz = (a00 * P_smooth) > P_smooth.max() * 1e-8
    a00 = a00[nz]
    
    # Main correction calculation
    P_norm = P_smooth[nz] / a00
    xK = window_xy[0] * kernel
    x2K = window_xy[0] * xK
    yK = window_xy[1] * kernel
    y2K = window_xy[1] * yK
    xyK = window_xy[1] * xK
    a10 = fftconvolve(mask, xK, mode='valid')[nz]
    a01 = fftconvolve(mask, yK, mode='valid')[nz]
    a11 = fftconvolve(mask, xyK, mode='valid')[nz]
    a20 = fftconvolve(mask, x2K, mode='valid')[nz]
    a02 = fftconvolve(mask, y2K, mode='valid')[nz]
    
    xP = fftconvolve(P_histogram, xK, mode='same')[nz]
    yP = fftconvolve(P_histogram, yK, mode='same')[nz]

    # Apply the correction, which is done inside the exponential
    # to keep everything positive.
    denom = (a20 * a01 ** 2 + a10 ** 2 * a02 - a00 * a02 * a20 + a11 ** 2 * a00 - 2 * a01 * a10 * a11)
    A = a11 ** 2 - a02 * a20
    Ax = a10 * a02 - a01 * a11
    Ay = a01 * a20 - a10 * a11
    scaling = (P_smooth[nz] * A + xP * Ax + yP * Ay) / denom
    
    P_final = P_smooth.copy()
    P_final[nz] = P_norm * np.exp(np.minimum(scaling / P_norm, 4) - 1)

    # Normalize and return
    P_final /= P_final.sum() * dx * dy
    return xmid, ymid, P_final



def make_1d_mask(N, window_width):
    full_width = N + 2 * window_width
    mask = np.ones(full_width)
    mask[:window_width] = 0
    mask[window_width] = 0.5
    mask[-window_width:] = 0
    mask[-(window_width+1)] = 0.5
    return mask




class Hints(object):
    def __init__(self):
        self._peak = None
        self._cov = None

    def has_peak(self):
        return self._peak is not None
    def set_peak(self, peak):
        self._peak = peak
    def get_peak(self):
        return self._peak
    def del_peak(self):
        self._peak = None

    def has_cov(self):
        return self._cov is not None
    def set_cov(self, cov):
        self._cov = cov
    def get_cov(self):
        return self._cov
    def del_cov(self):
        self._cov = None
    
    def update(self, other):
        if other.has_peak():
            self.set_peak(other.get_peak())
        if other.has_cov():
            self.set_cov(other.get_cov())


import itertools
import numpy as np
from cosmosis.output.text_output import TextColumnOutput
from .. import ParallelSampler


def task(p):
    i,p = p
    results = list_sampler.pipeline.run_results(p, all_params=True)
    #If requested, save the data to file
    if list_sampler.save_name and results.block is not None:
        results.block.save_to_file(list_sampler.save_name+"_%d"%i, clobber=True)
    return results.post, (results.prior, results.extra)




class ListSampler(ParallelSampler):
    parallel_output = False
    sampler_outputs = [("prior", float), ("post", float)]

    def config(self):
        global list_sampler
        list_sampler = self

        self.converged = False
        self.filename = self.read_ini("filename", str)
        self.save_name = self.read_ini("save", str, "")
        self.burn = self.read_ini("burn", int, 0)
        self.thin = self.read_ini("thin", int, 1)
        limits = self.read_ini("limits", bool, False)

        #overwrite the parameter limits
        if not limits:
            if self.output is not None:
                self.output.columns = []
            for p in self.pipeline.parameters:
                p.limits = (-np.inf, np.inf)
                if self.output is not None:
                    self.output.add_column(str(p), float)
            if self.output is not None:
                for p in self.pipeline.extra_saves:
                    self.output.add_column('{}--{}'.format(*p), float)
                for p,ptype in self.sampler_outputs:
                    self.output.add_column(p, ptype)


    def execute(self):

        #Load in the filename that was requested
        file_options = {"filename":self.filename}
        column_names, samples, _, _, _ = TextColumnOutput.load_from_options(file_options)
        samples = samples[0]
        # find where in the parameter vector of the pipeline
        # each of the table parameters can be found
        replaced_params = []
        for i,column_name in enumerate(column_names):
            # ignore additional columns like e.g. "like", "weight"
            try:
                section,name = column_name.split('--')
            except ValueError:
                print("Not including column %s as not a cosmosis name" % column_name)
                continue
            section = section.lower()
            name = name.lower()
            # find the parameter in the pipeline parameter vector
            # may not be in there - warn about this
            try:
                j = self.pipeline.parameters.index((section,name))
                replaced_params.append((i,j))
            except ValueError:
                print("Not including column %s as not in values file" % column_name)

        #Create a collection of sample vectors at the start position.
        #This has to be a list, not an array, as it can contain integer parameters,
        #unlike most samplers
        v0 = self.pipeline.start_vector(all_params=True, as_array=False)
        sample_vectors = [v0[:] for i in range(len(samples))]
        #Fill in the varied parameters. We are not using the
        #standard parameter vector in the pipeline with its 
        #split according to the ini file
        for s, v in zip(samples, sample_vectors):
            for i,j in replaced_params:
                v[j] = s[i]

        #Turn this into a list of jobs to be run 
        #by the function above
        sample_index = list(range(len(sample_vectors)))
        jobs = list(zip(sample_index, sample_vectors))

        #Run all the parameters
        #This only outputs them all at the end
        #which is a bit problematic, though you 
        #can't use MPI and retain the output ordering.
        #Have a few options depending on whether
        #you care about this we should think about
        #(also true for grid sampler).
        if self.pool:
            results = self.pool.map(task, jobs)
        else:
            results = list(map(task, jobs))

        #Save the results of the sampling
        #We now need to abuse the output code a little.
        for sample, result  in zip(sample_vectors, results):
            #Optionally save all the results calculated by each
            #pipeline run to files
            (prob, (prior,extra)) = result
            #always save the usual text output
            self.output.parameters(sample, extra, prior, prob)
        #We only ever run this once, though that could 
        #change if we decide to split up the runs
        self.converged = True

    def is_converged(self):
        return self.converged



#coding: utf-8


u"""Definition of the :class:`Inifile` class."""


import os
import sys
import collections
import warnings
import configparser

class CosmosisConfigurationError(configparser.Error):
    u"""Something to throw when there is an error in a .ini file particular to Cosmosis.

    The (underlying) object simply carries a string providing information
    to the user.

    """
    pass

class IncludingConfigParser(configparser.ConfigParser):
    u"""Extension of built-in python :class:`ConfigParser` to \%include other files.

    Use the line: %include filename.ini This is assumed to end a section,
    and the last section in the included file is assumed to end as well

    Note that the caller of the `read()` method may set a
    `no_expand_includes` attribute on this object, to cause any %include
    lines to *not* actually be actioned (they will be regarded as
    comments, but still delineate sections).
    """

    def __init__(self, defaults=None, print_include_messages=True):
        configparser.ConfigParser.__init__(self,
                                   defaults=defaults,
                                   dict_type=collections.OrderedDict,
                                   strict=False)
        self.print_include_messages = print_include_messages

    def _read(self, fp, fpname):
        u"""Parse a sectioned setup file.

        The sections in setup file contains a title line at the top,
        indicated by a name in square brackets (`[]'), plus key/value
        options lines, indicated by `name: value' format lines.
        Continuations are represented by an embedded newline then
        leading whitespace.  Blank lines, lines beginning with a '#',
        and just about everything else are ignored.

        """
        cursect = None                        # None, or a dictionary
        optname = None
        lineno = 0
        e = None                              # None, or an exception
        while True:
            line = fp.readline()
            if not line:
                break
            lineno = lineno + 1
            # comment or blank line?
            if line.strip() == '' or line[0] in '#;':
                continue
            if line.split(None, 1)[0].lower() == 'rem' and line[0] in "rR":
                # no leading whitespace
                continue
            # continuation line?
            if line[0].isspace() and cursect is not None and optname:
                value = line.strip()
                if value:
                    cursect[optname].append(value)
            # a section header or option header?
            else:
                #JAZ add environment variable expansion
                if not getattr(self, 'no_expand_vars', False):
                    line = os.path.expandvars(line)
                # is it a section header?
                mo = self.SECTCRE.match(line)
                if mo:
                    sectname = mo.group('header')
                    if sectname in self._sections:
                        cursect = self._sections[sectname]
                    elif sectname == configparser.DEFAULTSECT:
                        cursect = self._defaults
                    else:
                        cursect = self._dict()
                        self._sections[sectname] = cursect
                    # So sections can't start with a continuation line
                    optname = None
                # no section header in the file?
                elif line.lower().startswith('%include'):
                    if  not  getattr (self, 'no_expand_includes', False):
                        include_statement, filename = line.split()
                        filename = filename.strip('"').strip("'")
                        if self.print_include_messages:
                            sys.stdout.write("Reading included ini file: `"
                                                               + filename + "'\n")
                        if not os.path.exists(filename):
                            # TODO: remove direct sys.stderr writes
                            raise ValueError("Tried to include non-existent "
                                          "ini file: `" + filename + "'\n")
                        self.read(filename)
                    cursect = None
                elif cursect is None:
                    raise configparser.MissingSectionHeaderError(fpname,
                                                                 lineno,
                                                                 line)
                # an option line?
                else:
                    mo = self._optcre.match(line)
                    if mo:
                        optname, vi, optval = mo.group('option', 'vi', 'value')
                        optname = self.optionxform(optname.rstrip())
                        # This check is fine because the OPTCRE cannot
                        # match if it would set optval to None
                        if optval is not None:
                            if vi in ('=', ':') and ';' in optval:
                                # ';' is a comment delimiter only if it
                                # follows a spacing character
                                pos = optval.find(';')
                                if pos != -1 and optval[pos-1].isspace():
                                    optval = optval[:pos]
                            optval = optval.strip()
                            # allow empty values
                            if optval == '""':
                                optval = ''
                            cursect[optname] = [optval]
                        else:
                            # valueless option handling
                            cursect[optname] = optval
                    else:
                        # a non-fatal parsing error occurred.  set up the
                        # exception but keep going. the exception will be
                        # raised at the end of the file and will contain
                        # a list of all bogus lines
                        if not e:
                            e = configparser.ParsingError(fpname)
                        e.append(lineno, repr(line))
        # if any parsing errors occurred, raise an exception
        if e:
            raise e

        # join the multi-line values collected while reading
        all_sections = [self._defaults]
        all_sections.extend(list(self._sections.values()))
        for options in all_sections:
            for name, val in list(options.items()):
                if isinstance(val, list):
                    options[name] = '\n'.join(val)




class Inifile(IncludingConfigParser):

    u"""A dictionary of `(section, name) -> value` pairs corresponding to entries in a .ini file.

    The class is designed to hide the details of parsing .ini files, and
    then for creating :class:`DataBlock` objects (which wrap C objects
    and can be passed down a processing pipeline which may include modules
    written in C) via the :class:`Pipeline` and then :class:`Module`
    constructors (see `Pipeline.__init__()`).

    The values are all stored as strings, with methods provided to locate
    and interpret them as integers, floating-point numbers, booleans, or
    arrays of integers or floating-point numbers.

    """

    def __init__(self, filename, defaults=None, override=None, print_include_messages=True):
        u"""Read in a configuration from `filename`.

        The `defaults` will be applied if a parameter is not specified in
        the file (or \%included descendants), and the `override`s will be
        imposed on the regardless of whether those parameters have
        assigned values or not.

        Where supplied, `defaults` and `override` should be dictionary
        mappings of `(section, name) -> value`.

        """

        IncludingConfigParser.__init__(self,
                                       defaults=defaults,
                                       print_include_messages=print_include_messages)

        # if we are pased a dict, convert it to an inifile
        if isinstance(filename, dict):
            for section, values in filename.items():
                self.add_section(section)
                for key, value in values.items():
                    self.set(section, key, str(value))
        # default read behaviour is to ignore unreadable files which
        # is probably not what we want here
        elif filename is not None:
            if isinstance(filename,str) and not os.path.exists(filename):
                raise IOError("Unable to open configuration file `" + filename + "'")
            self.read(filename)

        # override parameters
        if override:
            for section, name in override:
                if section=="DEFAULT":
                    self._defaults[name] = override[(section,name)]
                else:
                    if not self.has_section(section):
                        self.add_section(section)
                    self.set(section, name, override[(section, name)])



    def __iter__(self):
        u"""Iterate over all the parameters.

        The value of the iterator is `((section, name), value)`.

        """
        return (((section, name), value) for section in self.sections()
                for name, value in self.items(section))



    def items(self, section, raw=False, vars=None, defaults=True):
        u"""Return a list of pairs (key, value) from all the options in a given `section`.

        If raw is set, do not replace values which are set using the ini file 
        interpolation syntax %(name)s.

        If vars is set to a dictionary, use it as an additional source of options.

        If defaults is True (the default), parameters in the [DEFAULT] section are included
        in all other sections.


        """
        if defaults:
            return IncludingConfigParser.items(self, section, raw=raw, vars=vars)
        else:
            d = collections.OrderedDict()
            try:
                d.update(self._sections[section])
            except KeyError:
                if section != configparser.DEFAULTSECT:
                    raise configparser.NoSectionError(section)
            # Update with the entry specific variables
            if vars:
                for key, value in list(vars.items()):
                    d[self.optionxform(key)] = value
            options = list(d.keys())
            if raw:
                return [(option, d[option])
                        for option in options]
            else:
                return [(option, self._interpolate(section, option, d[option], d))
                        for option in options]


    def get(self, section, option, raw=False, vars=None, fallback=configparser._UNSET):

        u"""Get a value as a string, or `default` if the value is not in the dictionary.

        If the `default` is not set and is needed, an error with a message
        to the user will be raised. (`None` is *not* acceptable as a
        default).

        """
        try:
            return IncludingConfigParser.get(self, section, option, raw=raw, vars=vars, fallback=fallback)
        except (configparser.NoSectionError, configparser.NoOptionError) as e:
            if fallback is configparser._UNSET:
                raise CosmosisConfigurationError("CosmoSIS looked for an option called '%s' in the '[%s]' section, but it was not in the ini file"%(option,section))
            else:
                return fallback

    # these functions override the default parsers to allow for extra formats
    def getint(self, section, option, raw=False, vars=None, fallback=configparser._UNSET):
        u"""Get a value as an integer, or return `default` if the value is not found.
        
        If the `default` is not set and is needed, an error with a message
        to the user will be raised. (`None` is *not* acceptable as a
        default).

        """
        try:
            return IncludingConfigParser.getint(self, section, option, raw=raw, vars=vars, fallback=fallback)
        except (configparser.NoSectionError, configparser.NoOptionError, CosmosisConfigurationError) as e:
            if fallback is configparser._UNSET:
                raise CosmosisConfigurationError("CosmoSIS looked for an integer option called '%s' in the '[%s]' section, but it was not in the ini file"%(option,section))
            elif not isinstance(fallback, int):
                raise TypeError("Default not integer")
            else:
                return fallback

    def getfloat(self, section, option, raw=False, vars=None, fallback=configparser._UNSET):
        u"""Get a floating-point value from the dictionary, with `default`.

        If the value is not found in the dictionary and `default` is
        specified, then `default` will be returned.  Otherwise an error
        will be thrown with a useful message for the user.

        """
        try:
            return IncludingConfigParser.getfloat(self, section, option, raw=raw, vars=vars, fallback=fallback)
        except (configparser.NoSectionError, configparser.NoOptionError, CosmosisConfigurationError) as e:
            if fallback is configparser._UNSET:
                raise CosmosisConfigurationError("CosmoSIS looked for a float option called '%s' in the '[%s]' section, but it was not in the ini file"%(option,section))
            elif not isinstance(fallback, float):
                raise TypeError("Default not float")
            else:
                return fallback

    def getboolean(self, section, option, raw=False, vars=None, fallback=configparser._UNSET):
        u"""Interpret a parameter as a boolean, including symbolic values.

        This essentially allows a configuration file to represent boolean
        values in the most convenient manner (‘true’, ‘n’, etc) as well as
        with zero/non-zero numerical values.

        If the parameter is not found in the dictionary, then `default`
        will be returned, which will itself default to `False` if not
        specified.

        """
        try:
            return IncludingConfigParser.getboolean(self, section, option, raw=raw, vars=vars, fallback=fallback)
        except ValueError:
            # additional options t/y/n/f
            value = self.get(section, option).lower()
            if value in ['y', 'yes', 't','true']:
                return True
            elif value in ['n', 'no', 'f', 'false']:
                return False
            else:
                raise ValueError("Unable to parse parameter "
                                 "%s--%s = %s into boolean form"
                                 % (section, option, value))
        except (configparser.NoSectionError, configparser.NoOptionError, CosmosisConfigurationError) as e:
            if fallback is configparser._UNSET:
                raise CosmosisConfigurationError("CosmoSIS looked for a boolean (T/F) option called '%s' in the '[%s]' section, but it was not in the ini file"%(option,section))
            elif not isinstance(fallback, bool):
                raise TypeError("Default not boolean")
            else:
                return fallback



    def gettyped(self, section, name):
        u"""Best-guess the type of a parameter and return it as that type.

        The method will try parsing the value as, in this order:
            an integer or list of integers,
            a float or list of floats,
            a complex number or list of complex numbers,
            a boolean,
            a string.

        The string value is the fallback if all else fails.
        """

        import re

        value = IncludingConfigParser.get(self, section, name)
        value = value.strip()
        # There isn't really a sensible thing to return for this,
        # so we just need to set it to None.
        if not value: return None
        # try quoted string
        m = re.match(r"^(['\"])(.*?)\1$", value)
        if m is not None:
            return m.group(2)

        value_list = value.split()

        # Try to match integer array.  This will fail whenever a decimal
        # point occurs anywhere in the list of values.
        try:
            parsed = [int(s) for s in value_list]
            if len(parsed) == 1:
                return parsed[0]
            else:
                return parsed
        except ValueError:
            pass

        # try to match float array
        try:
            parsed = [float(s) for s in value_list]
            if len(parsed) == 1:
                return parsed[0]
            else:
                return parsed
        except ValueError:
            pass

        # try to match complex array
        try:
            parsed = [complex(s) for s in value_list]
            if len(parsed) == 1:
                return parsed[0]
            else:
                return parsed
        except ValueError:
            pass

        # try to match boolean (no array support)

        try:
            return self.getboolean(section, name)
        except ValueError:
            pass

        # default to string
        return value



from cosmosis.datablock.cosmosis_py import DataBlock
import cosmosis.datablock.cosmosis_py.errors as errors
import numpy as np
import tempfile
import os
import yaml
import pytest

def test_string_array():
    block = DataBlock()
    # 1 input as list
    strings = "I am a mole and I live in a hole".split()
    print(strings)
    block.put_string_array_1d("my_section", "my_key", strings)
    strings2 = block.get_string_array_1d("my_section", "my_key")
    print(strings2)

    assert (strings == strings2).all()
    # 2 input as array
    block.put_string_array_1d("my_section2", "my_key2", np.array(strings))
    strings3 = block.get_string_array_1d("my_section2", "my_key2")

    assert (strings == strings3).all()
    print(strings3)


def test_string_array_getset():
    block = DataBlock()
    # 1 input as list
    strings = "I am a mole and I live in a hole".split()
    block["my_section", "my_key"] = strings

    strings2 = block.get_string_array_1d("my_section", "my_key")
    assert (strings == strings2).all()

    strings3 = block["my_section", "my_key"]


def test_string_array_save():
    block = DataBlock()
    # 1 input as list
    strings = "I am a mole and I live in a hole".split()
    block["my_section", "my_key"] = strings

    with tempfile.TemporaryDirectory() as dirname:
        dn = os.path.join(dirname, "dir")
        fn = os.path.join(dirname, "tar")
        yn = os.path.join(dirname, "block.yml")
        block.save_to_directory(dn)
        block.save_to_directory(fn)
        block.to_yaml(yn)

        strings2 = np.loadtxt(f"{dn}/my_section/my_key.txt", dtype=str)
        assert (strings == strings2).all()

        strings3 = yaml.safe_load(open(yn))["my_section"]["my_key"]
        assert (strings == strings3)


def test_delete():
    b = DataBlock()
    section = 'test'
    b.put_bool(section, 'b', True)
    b._delete_section(section)
    assert len(list(b.keys()))==0
    with pytest.raises(errors.BlockSectionNotFound):
        b._delete_section(section)

def test_bool():
    b = DataBlock()
    section = 'test'
    b.put_bool(section, 'b', True)
    assert b.get_bool(section, 'b')==True
    assert b.get_bool(section, 'c', default=False)==False
    assert b.get_bool(section, 'd', default=True)==True
    with pytest.raises(errors.BlockSectionNotFound):
        b.get_bool('ddd', 'c')

def test_default():
    b = DataBlock()
    section = 'test'
    assert b.get_int(section, 'a', default=14)==14
    with pytest.raises(errors.BlockSectionNotFound):
        b.get_int('test2', 'a')

    assert b.get_int(section, 'a')==14
    assert b[section, 'a'] == 14
    assert b.get_double(section, 'b', default=1.0)==1.0
    assert b.get_string(section, 'c', default="QQQ")=="QQQ"
    assert b.get_string(section, 'c', default="VVV")=="QQQ"

def test_complex():
    pass

def test_int():
    # make sure the shuffled sequence does not lose any elements
    b = DataBlock()
    section = 'test'
    b.put_int(section,'a',1)
    assert b.get_int(section,'a')==1
    with pytest.raises(errors.BlockNameNotFound):
        b.get_int('test', 'b')

    b.replace_int(section,'a',2)

    assert b.get_int(section,'a')==2
    with pytest.raises(errors.BlockNameAlreadyExists):
        b.put_int('test', 'a', 3)


def test_string():
    b = DataBlock()
    section = 'test'
    b.put_string(section, 's', 'my_string')
    assert b.get_string(section,'s')=='my_string'
    with pytest.raises(errors.BlockNameNotFound):
        b.get_string('test', 't')
    with pytest.raises(errors.BlockNameAlreadyExists):
        b.put_string('test', 's', 'my_string')


def test_int_array():
    b = DataBlock()
    section='test'
    b.put_int_array_1d(section, 'x', [1,2,3])
    r = b.get_int_array_1d(section, 'x')
    assert (r==[1,2,3]).all()

def test_double_array():
    b = DataBlock()
    section='test'
    b.put_double_array_1d(section, 'x', [1.4,2.1,3.6])
    r = b.get_double_array_1d(section, 'x')
    assert (r==[1.4,2.1,3.6]).all() 

def test_keys():
    b = DataBlock()
    section='dogs'
    b.put(section, 'x', [1.4,2.1,3.6])
    b.put(section, "n", 14)
    b.put(section, 's', 'my_string')
    section='other'
    b.put(section, 'a', 98)
    b.put(section, "b", 1.4)
    b.put_string(section, 's', 'my_string')
    keys = list(b.keys())
    assert sorted(keys) == sorted([('dogs','x'), ('dogs','n'),('dogs','s'),('other','a'),('other','b'),('other','s')])
    for k in keys:
        assert k in b




if __name__ == '__main__':
    # test_string_array()
    test_string_array_save()    


#coding: utf-8
from .. import ParallelSampler
import ctypes as ct
import os
import cosmosis
import numpy as np
import sys

loglike_type = ct.CFUNCTYPE(ct.c_double, 
    ct.POINTER(ct.c_double),  #cube
    ct.c_int,   #ndim
    ct.c_int,   #npars
    ct.c_voidp, #context
)


dumper_type = ct.CFUNCTYPE(None, #void
    ct.c_int,  #nsamples
    ct.c_int,  #nlive
    ct.c_int,  #npars
    ct.POINTER(ct.c_double),   #physlive
    ct.POINTER(ct.c_double),   #posterior
    ct.POINTER(ct.c_double),   #paramConstr
    ct.c_double,   #maxLogLike
    ct.c_double,   #logZ
    ct.c_double,   #INSLogZ
    ct.c_double,   #logZerr
    ct.c_voidp,   #Context
)


multinest_args = [
    ct.c_bool,  #nest_IS   nested importance sampling 
    ct.c_bool,  #nest_mmodal   mode separation
    ct.c_bool,  #nest_ceff   constant efficiency mode
    ct.c_int,   #nest_nlive
    ct.c_double, #nest_tol
    ct.c_double, #nest_ef
    ct.c_int, #nest_ndims,
    ct.c_int, #nest_totPar,
    ct.c_int, #nest_nCdims,
    ct.c_int, #maxClst,
    ct.c_int, #nest_updInt,
    ct.c_double, #nest_Ztol,
    ct.c_char_p, #nest_root,
    ct.c_int, #seed,
    ct.POINTER(ct.c_int), #nest_pWrap,
    ct.c_bool, #nest_fb,
    ct.c_bool, #nest_resume,
    ct.c_bool, #nest_outfile,
    ct.c_bool, #initMPI,
    ct.c_double, #nest_logZero,
    ct.c_int, #nest_maxIter,
    loglike_type, #loglike,
    dumper_type, #dumper,
    ct.c_voidp, #context
]


MULTINEST_SECTION='multinest'


class MultinestSampler(ParallelSampler):
    parallel_output = False
    sampler_outputs = [("prior", float), ("like", float), ("post", float), ("weight", float)]
    supports_smp=False

    def config(self):
        if self.pool:
            libname = "libnest3_mpi.so"
        else:
            libname = "libnest3.so"

        dirname = os.path.split(__file__)[0]
        libname = os.path.join(dirname, "multinest_src", libname)
            
        try:
            libnest3 = ct.cdll.LoadLibrary(libname)
        except Exception as error:
            raise RuntimeError("Multinest could not be loaded.\n"
                             "This may mean an MPI compiler was not found to compile it,\n"
                             "or that some other error occurred.  More info below:\n"
                             + str(error)+'\n')

        self._run = libnest3.run
        self._run.restype=None
        self._run.argtypes = multinest_args
        self.converged=False

        self.ndim = len(self.pipeline.varied_params)

        # We add one to the output to save the posterior as well as the
        # likelihood.
        self.npar = self.ndim + len(self.pipeline.extra_saves) + 2

        #Required options
        self.max_iterations = self.read_ini("max_iterations", int)
        self.live_points    = self.read_ini("live_points", int)

        #Output and feedback options
        self.feedback               = self.read_ini("feedback", bool, True)
        self.resume                 = self.read_ini("resume", bool, False)
        self.multinest_outfile_root = self.read_ini("multinest_outfile_root", str, "")
        self.update_interval        = self.read_ini("update_interval", int, 200)

        if len(self.multinest_outfile_root)>511:
            raise ValueError("Multinest parameter 'multinest_outfile_root'"
                " cannot be longer than 511 characters.")
        #General run options
        self.random_seed = self.read_ini("random_seed", int, -1)
        self.importance  = self.read_ini("ins", bool, True)
        self.efficiency  = self.read_ini("efficiency", float, 1.0)
        self.tolerance   = self.read_ini("tolerance", float, 0.1)
        self.log_zero    = self.read_ini("log_zero", float, -1e6)

        #Multi-modal options
        self.mode_separation    = self.read_ini("mode_separation", bool, False)
        self.const_efficiency   = self.read_ini("constant_efficiency", bool, False)
        self.max_modes          = self.read_ini("max_modes", int, default=100)
        self.cluster_dimensions = self.read_ini("cluster_dimensions", int, default=-1)
        self.mode_ztolerance    = self.read_ini("mode_ztolerance", float, default=0.5)

        #Parameters with wrap-around edges - can help sampling
        #of parameters which are relatively flat in likelihood
        wrapped_params = self.read_ini("wrapped_params", str, default="")
        wrapped_params = wrapped_params.split()
        self.wrapping = [0 for i in range(self.ndim)]
        if wrapped_params:
            print("")
        for p in wrapped_params:
            try:
                P = p.split('--')
            except ValueError:
                raise ValueError("You included {} in wrapped_params mulitnest option but should be format: section--name".format(p))
            if P in self.pipeline.varied_params:
                index = self.pipeline.varied_params.index(P)
                self.wrapping[index] = 1
                print("MULTINEST: Parameter {} ({}) will be wrapped around the edge of its prior".format(index,p))
            elif P in self.pipeline.parameters:
                print("MULTINEST NOTE: You asked for wrapped sampling on {}. That parameter is not fixed in this pipeline, so this will have no effect.".format(p))
            else:
                raise ValueError("You asked for an unknown parameter, {} to be wrapped around in the multinest wrapped_params option.".format(p))
        if wrapped_params:
            print("")



        if self.output:
            def dumper(nsample, nlive, nparam, live, post, paramConstr, max_log_like, logz, ins_logz, log_z_err, context):
                print("Saving %d samples" % nsample)
                self.output_params(nsample, live, post, logz, ins_logz, log_z_err)
            self.wrapped_output_logger = dumper_type(dumper)
        else:
            def dumper(nsample, nlive, nparam, live, post, paramConstr, max_log_like, logz, ins_logz, log_z_err, context):
                return
            self.wrapped_output_logger = dumper_type(dumper)

        def likelihood(cube_p, ndim, nparam, context_p):
            # The -1 is because we store the likelihood separately.
            nextra = nparam-ndim-2
            #pull out values from cube
            cube_vector = np.array([cube_p[i] for i in range(ndim)])
            vector = self.pipeline.denormalize_vector_from_prior(cube_vector)

            try:
                results = self.pipeline.run_results(vector)
            except KeyboardInterrupt:
                raise sys.exit(1)

            for i in range(ndim):
                cube_p[i] = vector[i]

            for i in range(nextra):
                cube_p[ndim+i] = results.extra[i]

            # posterior column
            cube_p[ndim+nextra] = results.prior
            cube_p[ndim+nextra+1] = results.post

            return results.like
        self.wrapped_likelihood = loglike_type(likelihood)

    def worker(self):
        self.sample()

    def execute(self):
        self.log_z = 0.0
        self.log_z_err = 0.0

        self.sample()

        self.output.final("log_z", self.log_z)
        self.output.final("log_z_error", self.log_z_err)

    def sample(self):
        # only master gets dumper function
        cluster_dimensions = self.ndim if self.cluster_dimensions==-1 else self.cluster_dimensions
        periodic_boundaries = (ct.c_int*self.ndim)()
        for i in range(self.ndim):
            periodic_boundaries[i] = self.wrapping[i]
        context=None
        init_mpi=False
        
        self._run(self.importance, self.mode_separation,
                  self.const_efficiency, self.live_points,
                  self.tolerance, self.efficiency, self.ndim,
                  self.npar, cluster_dimensions, self.max_modes,
                  self.update_interval, self.mode_ztolerance,
                  self.multinest_outfile_root.encode('ascii'), self.random_seed,
                  periodic_boundaries, self.feedback, self.resume,
                  self.multinest_outfile_root!="", init_mpi,
                  self.log_zero, self.max_iterations, 
                  self.wrapped_likelihood, self.wrapped_output_logger,
                  context)

        self.converged = True

    def output_params(self, n, live, posterior, log_z, ins_log_z, log_z_err):
        self.output.reset_to_chain_start()
        self.log_z = ins_log_z if self.importance else log_z
        self.log_z_err = log_z_err
        data = np.array([posterior[i] for i in range(n*(self.npar+2))]).reshape((self.npar+2, n))
        for row in data.T:
            params = row[:self.ndim]
            extra_vals = row[self.ndim:self.npar-2]
            prior = row[self.npar-2]
            post = row[self.npar-1]
            like = row[self.npar]
            importance = row[self.npar+1]
            self.output.parameters(params, extra_vals, prior, like, post, importance)
        self.output.final("nsample", n)
        self.output.flush()

    def is_converged(self):
        return self.converged



import scipy.interpolate
import scipy.integrate
import numpy as np
from .datablock import names, SectionOptions
from .runtime import FunctionModule
import traceback 

try:
    from packaging.version import parse as parse_version
    min_hermitian_version = parse_version('1.17.0')
    use_hermitian_keyword = parse_version(np.version.version) > min_hermitian_version
except ImportError: # fall back to not using hermitian
    use_hermitian_keyword = False


MISSING = "if_you_see_this_there_was_a_mistake_creating_a_gaussian_likelihood"


class GaussianLikelihood:
    """
    Gaussian likelihood with a fixed covariance.  
    
    Subclasses must override build_data and build_covariance,
    e.g. to load from file.
    """
    x_section = MISSING
    x_name    = MISSING
    y_section = MISSING
    y_name    = MISSING
    like_name = MISSING

    #Set this to False to load the covariance at 
    #each cosmology instead of once at the start
    constant_covariance = True

    def __init__(self, options):
        self.options=options
        self.data_x, self.data_y = self.build_data()
        self.likelihood_only = options.get_bool('likelihood_only', False)

        if self.constant_covariance:
            self.cov = self.build_covariance()
            self.inv_cov = self.build_inverse_covariance()

            if not self.likelihood_only:
                self.chol = np.linalg.cholesky(self.cov)


            # We may want to include the normalization of the likelihood
            # via the log |C| term.
            include_norm = self.options.get_bool("include_norm", False)
            if include_norm:
                # We have no datablock here so we don't want to call any subclass method
                self.log_det_constant = GaussianLikelihood.extract_covariance_log_determinant(self,None)
                print("Including -0.5*|C| normalization in {} likelihood where |C| = {}".format(self.like_name, self.log_det_constant))
            else:
                self.log_det_constant = 0.0

        #Interpolation type, when interpolating into theory vectors
        self.kind = self.options.get_string("kind", "cubic")

        #Allow over-riding where the inputs come from in 
        #the options section
        if options.has_value("x_section"):
            self.x_section = options['x_section']
        if options.has_value("y_section"):
            self.y_section = options['y_section']
        if options.has_value("x_name"):
            self.x_name = options['x_name']
        if options.has_value("y_name"):
            self.y_name = options['y_name']
        if options.has_value("like_name"):
            self.like_name = options['like_name']



    def build_data(self):
        """
        Override the build_data method to read or generate 
        the observed data vector to be used in the likelihood
        """
        raise RuntimeError("Your Gaussian covariance code needs to "
            "over-ride the build_data method so it knows how to "
            "load the observed data")
        #using info in self.options,
        #like filenames etc,
        #build x to which we must interpolate
        #return x, y
    
    def build_covariance(self):
        """
        Override the build_covariance method to read or generate 
        the observed covariance
        """
        raise RuntimeError("Your Gaussian covariance code needs to "
            "over-ride the build_covariance method so it knows how to "
            "load the data covariance (or set constant_covariance=False and "
            "over-ride the extract_covariance method)")

        #using info in self.options,
        #like filenames etc,
        #build covariance

    def build_inverse_covariance(self):
        """
        Override the build_inverse_covariance method to change
        how the inverse is generated from the covariance.

        When the covariance is generated from a suite of simulations,
        for example, the simple inverse is not the best estimate.

        """
        # inverse of symmetric matrix should remain symmetric
        if np.allclose(self.cov, self.cov.T) and use_hermitian_keyword:
            return np.linalg.pinv(self.cov, hermitian=True)
        return np.linalg.inv(self.cov)


    def cleanup(self):
        """
        You can override the cleanup method if you do something 
        unusual to get your data, like open a database or something.
        It is run just once, at the end of the pipeline.
        """
        pass
    
    def extract_covariance(self, block):
        """
        Override this and set constant_covariance=False
        to enable a cosmology-dependent covariance.

        Load the covariance from the block here.
        """
        raise RuntimeError("You need to implement the method "
            "'extract_covariance' if you set constant_covariance=False "
            "in a gaussian likelihood")

    def extract_inverse_covariance(self, block):
        """
        Override this and set constant_covariance=False
        to enable a cosmology-dependent inverse
        covariance matrix.

        By default the inverse is just directly calculated from
        the covariance, but you might have a different method.
        """
        return np.linalg.inv(self.cov)

    def extract_covariance_log_determinant(self, block):
        """
        If you are using a varying covariance we have to account
        for the dependence of the covariance matrix on parameters
        in the likelihood.

        Override this method if you have a faster way to get |C|
        rather than just taking the determinant of directly (e.g.
        if you know it is diagonal or block-diagonal).

        Since we know that we must have the inverse covariance,
        whereas the covariance itself is optional, we use the former
        in the default implementation.
        """
        sign, log_inv_det = np.linalg.slogdet(self.inv_cov)
        log_det = -log_inv_det
        return log_det


    def do_likelihood(self, block):
        #get data x by interpolation
        x = np.atleast_1d(self.extract_theory_points(block))
        mu = np.atleast_1d(self.data_y)

        #If covariance is a function of parameters, compute the 
        #new one now.
        if not self.constant_covariance:
            self.cov = np.atleast_2d(self.extract_covariance(block))
            self.inv_cov = np.atleast_2d(self.extract_inverse_covariance(block))

        #gaussian likelihood
        d = x-mu
        chi2 = np.einsum('i,ij,j', d, self.inv_cov, d)
        chi2 = float(chi2)
        like = -0.5*chi2

        #It can be useful to save the chi^2 as well as the likelihood,
        #especially when the covariance is non-constant.
        block[names.data_vector, self.like_name+"_CHI2"] = chi2

        #if the covariance is a function of parameters then we must 
        #account for this in the likelihood.
        if not self.constant_covariance:
            log_det = self.extract_covariance_log_determinant(block)
        else:
            log_det = self.log_det_constant            

        norm = -0.5 * log_det
        like += norm
        block[names.data_vector, self.like_name+"_LOG_DET"] = float(log_det)
        block[names.data_vector, self.like_name+"_NORM"] = float(norm)

        # Numpy has started returning a 0D array in recent versions (1.14).
        # Convert this to a float.
        like = float(like)

        #Now save the resulting likelihood
        block[names.likelihoods, self.like_name+"_LIKE"] = like

        # For some very fast likelihoods the overhead from
        # the steps below is painful.  Setting likelihood_only avoids that.
        if self.likelihood_only:
            return

        #And also the predicted data points - the vector of observables 
        # that in a fisher approch we want the derivatives of.
        #and inverse cov mat which also goes into the fisher matrix.
        block[names.data_vector, self.like_name + "_theory"] = x
        block[names.data_vector, self.like_name + "_data"] = mu
        block[names.data_vector, self.like_name + "_inverse_covariance"] = self.inv_cov

        #We might just be calculating the inverse cov and ignoring the covmat.
        #in that case we do not try to save it
        if self.cov is not None:
            block[names.data_vector, self.like_name + "_covariance"] = self.cov
            #Also save a simulation of the data - the mean with added noise
            #these can be used among other places by the ABC sampler.
            #This also requires the cov mat.
            # If we have a parameter-dependent covariance we need
            # to re-calculate the Cholesky decomposition to simulate some data.
            if not self.constant_covariance:
                self.chol = np.linalg.cholesky(self.cov)
            sim = self.simulate_data_vector(x)
            block[names.data_vector, self.like_name + "_simulation"] = sim

    def simulate_data_vector(self, x):
        "Simulate a data vector by adding a realization of the covariance to the mean"
        #generate a vector of normal deviates

        r = np.random.randn(x.size)
        return x + np.dot(self.chol, r)


    def extract_theory_points(self, block):
        "Extract relevant theory from block and get theory at data x values"
        theory_x = block[self.x_section, self.x_name]
        theory_y = block[self.y_section, self.y_name]
        return self.generate_theory_points(theory_x, theory_y)

    def generate_theory_points(self, theory_x, theory_y):
        "Generate theory predicted data points by interpolation into the theory"
        f = scipy.interpolate.interp1d(theory_x, theory_y, kind=self.kind)
        return np.atleast_1d(f(self.data_x))

    @classmethod
    def build_module(cls):

        def setup(options):
            options = SectionOptions(options)
            likelihoodCalculator = cls(options)
            return likelihoodCalculator

        def execute(block, config):
            likelihoodCalculator = config
            likelihoodCalculator.do_likelihood(block)
            return 0

        def cleanup(config):
            likelihoodCalculator = config
            likelihoodCalculator.cleanup()

        return setup, execute, cleanup

    @classmethod
    def as_module(cls, name):
        setup, execute, cleanup = cls.build_module()
        return FunctionModule(name, setup, execute, cleanup)


class SingleValueGaussianLikelihood(GaussianLikelihood):
    """
    A Gaussian likelihood whos input is a single calculated value
    not a vector
    """
    name = MISSING
    section = MISSING
    like_name = MISSING
    mean = None
    sigma = None
    def __init__(self, options):
        self.options=options

        #First try getting the value from the class itself
        mean, sigma = self.build_data()

        if options.has_value("mean"):
            mean = options["mean"]
        if options.has_value("sigma"):
            sigma = options["sigma"]

        if sigma is None or mean is None:
            raise ValueError("Need to specify Gaussian mean/sigma for '{0}' \
                either in class definition, build_data method, or in the ini \
                file".format(self.like_name))
        if options.has_value("like_name"):
            self.like_name = options["like_name"]
        print('Likelihood "{0}" will be Gaussian {1} +/- {2} '.format(self.like_name, mean, sigma))
        self.data_y = np.array([mean])
        self.cov = np.array([[sigma**2]])
        self.inv_cov = np.array([[sigma**-2]])

        include_norm = self.options.get_bool("include_norm", False)
        if include_norm:
            # We have no datablock here so we don't want to call any subclass method
            self.log_det_constant = GaussianLikelihood.extract_covariance_log_determinant(self,None)
            print("Including -0.5*|C| normalization in {} likelihood where log|C| = {}".format(self.like_name, self.log_det_constant))
        else:
            self.log_det_constant = 0.0

        self.likelihood_only = options.get_bool('likelihood_only', 'False')
            
    def build_data(self):
        """Sub-classes can over-ride this if they wish, to generate 
        the data point in a more complex way"""
        return self.mean, self.sigma

    def build_covariance(self):
        """This method is only defined here to satisfy the superclass requirements. 
        There is no point over-riding it"""
        raise RuntimeError("Internal cosmosis error in SingleValueGaussianLikelihood")

    def extract_theory_points(self, block):
        "Extract relevant theory from block and get theory at data x values"
        return np.atleast_1d(block[self.section, self.name])



class WindowedGaussianLikelihood(GaussianLikelihood):
    def __init__(self, options):
        super(WindowedGaussianLikelihood, self).__init__(options)
        self.windows = self.build_windows()

    def build_windows(self):
        #This method should return a list of pairs of window_x, window_y:
        # [(x_1, w_1), (x_2, w_2), (x_3, w_3), ...]
        # There should be one window per data point.
        raise RuntimeError("When you set up a new windowed Gaussian likelihood you need to write the build_windows method")

    def generate_theory_points(self, theory_x, theory_y):
        "Generate theory predicted data points using window function"
        f = scipy.interpolate.interp(theory_x, theory_y, kind=self.kind)
        values = []
        for window_x, window_y in self.windows:
            xmin = max(window_x.min(), theory_x.min())
            xmax = min(window_x.max(), theory_x.max())
            w = scipy.interpolate.interp(window_x, window_y, kind=self.kind)
            g = lambda x: w(x)*f(x)
            v = scipy.integrate.romberg(g, xmin, xmax)
            values.append(v)
        return np.atleast_1d(values)
        




from .text_output import TextColumnOutput
from .cosmomc_output import CosmoMCOutput
from .null_output import NullOutput
from .fits_output import FitsOutput
from .in_memory_output import InMemoryOutput
from .output_base import output_registry
import logging


verbosity_levels = {
	"highest": 50-0,             #50
	"debug":50-logging.DEBUG,    #40
	"noisy":50-15,               #35
	"standard":50-logging.INFO,  #30
	"gentle":50-logging.WARNING, #20
	"quiet":50-logging.ERROR,    #10
	"silent":50-logging.FATAL,   #-1
}


def set_verbosity(verb):
	try:
		verb = int(verb)
	except ValueError:
		pass
	if not isinstance(verb, int):
		try:
			verb = verbosity_levels[verb]
		except KeyError:
			valid_levels = ', '.join(list(verbosity_levels.keys()))
			message = """Error specifiying verbosity.
				You put: '{0}'.
				We want either an integer 0 (silent) - 50 (everything) 
				or one of: {1}""".format(verb, valid_levels)
			raise ValueError(message)
	level = 50 - verb
	logging.getLogger().setLevel(level)
	logging.debug("CosmoSIS verbosity set to %d"%(verb))

def output_from_options(options, resume=False):
	# figure out the type of output required
	format = options.get('format', 'text')
	if format not in output_registry:
		known = '\n'.join('    %s - %s' % (f,output_registry[f].__name__) for f in output_registry)
		message = """
I do not know what format you meant by '%s' in the [output] ini file section.
I know about these format names:
%s
		""" % (format, known)
		print(message)
		raise KeyError("Unknown format")

	output_class = output_registry[format]

	verb = options.get("verbosity", "standard")
	set_verbosity(verb)
	return output_class.from_options(options,resume)

def input_from_options(options):
    format = options['format']
    if format not in output_registry:
        known = '\n'.join('    %s - %s' % (f,output_registry[f].__name__) for f in output_registry)
        message = """
I do not know what format you meant by '%s' in the [output] ini file section.
I know about these format names:
%s
        """ % (format, known)
        print(message)
        raise KeyError("Unknown format")

    output_class = output_registry[format]
    return output_class.load_from_options(options)

def test():
	from . import test_output
	test_output.test_text()



from . import lazy_pylab as pylab


class PostprocessProduct(object):
    def __init__(self, name, filename, value, info=None):
        self.name = name
        self.filename = filename
        self.value = value
        self.info = info

    def finalize(self):
        pass

class PostprocessPlot(PostprocessProduct):
    def finalize(self):
        pylab.figure(self.value.number)
        pylab.savefig(self.filename)
        pylab.close()

    def tweak(self, tweak):
        print("Tweaking", self.name)
        pylab.figure(self.value.number)
        tweak.info = self.info
        tweak.run()


class PostprocessText(PostprocessProduct):
    def finalize(self):
        self.value.close()



# cobaya imports matplotlib, which does
# some heavy duty output when logging is set to high.
# This actually reflects my lack of understanding of the
# python logging mechanism, and I should really go through
# and fix all of this, but for now we just temporarily suppress
# the debug info
import logging
logger = logging.getLogger()
level = logger.level
logger.setLevel(logging.WARNING)
from cobaya.samplers.mcmc.proposal import BlockedProposer
logger.setLevel(level)


# The cobaya proposal updates the parameter vector in-place,
# so we wrap it slightly to make a copy instead, as expected
# by the metropolis sampler
class CobayaProposalWrapper(BlockedProposer):
    def propose(self, p):
        q = p.copy()
        self.get_proposal(q)
        return q
    def propose_slow(self, p):
        q = p.copy()
        self.get_proposal_slow(q)
        return q
    def propose_fast(self, p):
        q = p.copy()
        self.get_proposal_fast(q)
        return q



from .text_output import TextColumnOutput
import numpy as np
import os
from glob import glob
from collections import OrderedDict
import numpy as np

PARAM_NAME = '.paramnames'


class CosmoMCOutput(TextColumnOutput):
    def __init__(self, filename, rank=0, nchain=1, delimiter='    ', lock=True, resume=False):
        super(CosmoMCOutput, self).__init__(filename, rank, nchain, '', lock=lock, resume=resume)
        if filename.endswith(self.FILE_EXTENSION):
            filename = filename[:-len(self.FILE_EXTENSION)]
        if rank == 0: 
            self._paramfile = open(filename+PARAM_NAME, 'w')
        else:
            self._paramfile = None
        self._last_params = None
        self._multiplicity = 0

    def _close(self):
        self._write_parameters_multiplicity()
        super(CosmoMCOutput, self)._close()
        if self._paramfile:
            self._paramfile.close()

    def _begun_sampling(self, params):
        if self._paramfile:
            print(self.columns)
            if self.columns[-1][0].upper() != "POST":
                raise RuntimeError("CosmoMC output format assumes "
                                       "likelihood is last column.")
            for c in self.columns[:-1]:
                self._paramfile.write(c[0]+'\n')
            self._paramfile.close()
        self._metadata=OrderedDict()

    def _write_comment(self, comment):
        #Do not think cosmomc can handle comments
        pass

    def _write_parameters(self, params):
        if (self._last_params is not None) and all([p==q for (p,q) in zip(self._last_params,params)]):
            self._multiplicity += 1
        else:
            self._write_parameters_multiplicity()
            self._last_params = params[:]
            self._multiplicity = 1
    
    def _write_parameters_multiplicity(self):
        if self._last_params:
            line = self.delimiter.join(('%16.7E'%x) for x
                                       in ([self._multiplicity] +
                                           self._last_params)) + '\n'
            self._file.write(line)

    @classmethod
    def load_from_options(cls, options):
        filename = options['filename']

        if filename.endswith(cls.FILE_EXTENSION):
            filename = filename[:-len(cls.FILE_EXTENSION)]

        # read column names from parameterfile
        column_names = [line.split()[0] for line in open(filename+PARAM_NAME)]
        column_names.append("LIKE")

        # first look for serial file
        if os.path.exists(filename+cls.FILE_EXTENSION):
            datafiles = [filename+cls.FILE_EXTENSION]
        else:
            datafiles = glob(filename+"_[0-9]*"+cls.FILE_EXTENSION)
            if not datafiles:
                raise RuntimeError("No datafiles found!")

        # cosmomc has no metadata support
        metadata = final_metadata = [{}]*len(datafiles)
        comments = []

        data = []
        for datafile in datafiles:
            print('LOADING CHAIN FROM FILE: ', datafile)
            chain = []
            with open(datafile) as f:
                for line in f:
                    vals = [float(word) for word in line.split()]
                    for i in range(int(vals[0])):
                        chain.append(vals[1:])
                chain = np.array(chain)
            print(datafile, chain.shape)
            data.append(chain)
        return column_names, data, metadata, comments, final_metadata

    @classmethod
    def from_options(cls, options, resume=False):
        if resume:
            raise ValueError("Cannot resume from cosmomc files yet")
        return super(CosmoMCOutput,cls).from_options(options, resume=False)




import os

TRUE_STRINGS =  ["T","t","True","TRUE","true","y","Y","yes","Yes","YES","1"]
FALSE_STRINGS = ["F","f","False","FALSE","false","n","N","no","No","NO","0"]

def boolean_string(s):
	if s is True or s is False:
		return s
	if not isinstance(s,str):
		raise ValueError("Non-string passed to boolean_string for conversion to bool")
	if s in TRUE_STRINGS:
		return True
	elif s in FALSE_STRINGS:
		return False
	raise ValueError("Could not convert string '%s' to True/False value"%s)


def try_numeric(x):
	if isinstance(x, int) or isinstance(x, float):
		return x
	try:
		return int(x)
	except:
		pass
	try:
		return float(x)
	except:
		pass
	return x

def parse_value(x):
    x = try_numeric(x)
    if x=='True':
        x=True
    if x=='False':
        x=False
    return x

def file_exists_and_is_empty(fpath):
    """
    
    from https://stackoverflow.com/questions/2507808/how-to-check-whether-a-file-is-empty-or-not
    """
    return os.path.isfile(fpath) and os.path.getsize(fpath) == 0

def file_exists_and_is_not_empty(fpath):
    """
    
    from https://stackoverflow.com/questions/2507808/how-to-check-whether-a-file-is-empty-or-not
    """
    return os.path.isfile(fpath) and os.path.getsize(fpath) > 0



#coding: utf-8


u"""Implementation of :class:`Prior` and the concrete probability distribution specializations.

The specializations are :class:`UniformPrior`, :class:`GaussianPrior`,
:class:`TruncatedGaussianPrior`, :class:`ExponentialPrior`,
:class:`TruncatedExponentialPrior`, :class:`TruncatedOneoverxPrior`,
and :class:`DeltaFunctionPrior`.

Applications can get by only knowing about the :class:`Prior` superclass.

"""

from . import config
import numpy as np
import math
from scipy import interpolate
import copy

class Prior(object):

    u"""This class serves as an abstract base for possible prior distributions, and provides a concrete distribution factory method.

    At high level, each concrete prior distribution is a function which
    takes a value and returns a probability of that value occurring;
    actually, the logarithm of the probability density is returned as this
    is more useful in a Bayesian framework where the value of evidence
    takes the form of a log-likelihood.  There are additionally methods
    which

    * Return a value corresponding to a given cumulated probability,
      i.e. the inverse of the prior distribution.

    * Obtain the value of a number of random samples from the
      distribution.

    * Optionally, a method for returning a truncated version of the
      distribution.

    The application calls a single function, :func:`Prior.load_priors`, to
    obtain a concrete set of all the priors the user specifies in the
    configuration files, and then abstractly calls the above methods on
    those priors to get the work done.

    """

    def __init__(self):
        u"""Do nothing."""
        pass



    def sample(self, n):
        u"""Return an array of `n` samples (default one) from the distribution in the derived object.

        The generic case implemented here in the base class is an
        *extremely* expensive operation in almost all cases.  Other cases
        (such as uniform and delta functions) are specially implemented to
        run much quicker.  Moral: think hard about how you are using this
        method.  If you have to deal with truncated distributions, maybe
        this sledge-hammer approach is the only one feasible.

        """
        if n is None:
            n = 1
        Y = np.random.uniform(0., 1.0, n)
        return np.array([self.denormalize_from_prior(y) for y in Y])



    @classmethod
    def parse_prior(cls, value):
        u"""Produce a concrete :class:`Prior` object based on a line of a .ini ifile.

        This is part of the implementation of :func:`load_priors` and
        should be considered private to the class.

        """
        prior_type, parameters = value.split(' ', 1)
        prior_type = prior_type.lower()
        try:
            # Most of the priors require float
            # parameters, but not all of them -
            # the InverseTransform prior needs
            # a string argument (file name).
            parameters_flt = []
            for p in parameters.split():
                try:
                    p = float(p)
                except:
                    pass
                parameters_flt.append(p)
            parameters = parameters_flt

            if prior_type.startswith("uni"):
                return UniformPrior(*parameters)
            elif prior_type.startswith("gau") or \
                    prior_type.startswith("nor"):
                return GaussianPrior(*parameters)
            elif prior_type.startswith("exp"):
                return ExponentialPrior(*parameters)
            elif  prior_type.startswith("one"):
                return TruncatedOneoverxPrior(*parameters)
            elif prior_type.startswith("tab") or \
                    prior_type.startswith("loa"):
                return TabulatedPDF(*parameters)
            else:
                raise ValueError("Unable to parse %s as prior" %
                                 (value,))
        except TypeError:
            raise ValueError("Unable to parse %s as prior" %
                             (value,))



    @classmethod
    def load_priors(cls,prior_files):
        u"""Produce a dictionary of `(section,name)` -> :class:`Prior` objects as per the instructions in `prior_files`.

        The dictionary values are concrete :class:`Prior` types, and this
        method is the applicationʼs sole constructor of :class:`Prior`
        objects.

        Can also pass in an ini file directly
        """
        priors = {}
        for f in prior_files:
            if isinstance(f, config.Inifile):
                ini = f
            else:
                ini = config.Inifile(f) 
            for option, value in ini:
                if option in priors:
                    raise ValueError("Duplicate prior identified")
                priors[option] = cls.parse_prior(value)

        return priors

class TabulatedPDF(Prior):

    u"""Load from a 2-column ASCII table containing values for x, pdf(x).
    
    pdf(x) not interpolated (assumed to be zero) outisde of range of x

    Heavily 'inspired' by DistDeviate from GalSim
    """

    def __init__(self, function_filename=None, lower=None, upper=None):
        u"""Create an object representing the distribution specified in function_filename"""
        

        self.function_filename = function_filename
        # Basic input checking and setups
        if function_filename is None:
            raise TypeError('You must specify a function_filename for TabulatedPDF!')

        # Load the probability distribution function, pdf(x)
        xarray, pdf = np.loadtxt(function_filename, unpack=True)
        
        if lower==None:
            self.lower = xarray.min()
        else:
            self.lower = lower

        if upper==None:
            self.upper = xarray.max()
        else:
            self.upper = upper

        pdf = pdf[(xarray >= self.lower)*(xarray <= self.upper)]
        xarray = xarray[(xarray >= self.lower)*(xarray <= self.upper)]

        # Set up pdf, so cumsum basically does a cumulative trapz integral
        # On Python 3.4, doing pdf[1:] += pdf[:-1] the last value gets messed up.
        # Writing it this way works.  (Maybe slightly slower though, so if we stop
        # supporting python 3.4, consider switching to the += version.)
        pdf_x = copy.copy(pdf)
        pdf[1:] = pdf[1:] + pdf[:-1]
        pdf[1:] *= np.diff(xarray)
        pdf[0] = 0

        # Check that the probability is nonnegative
        if not np.all(pdf >= 0.):
            raise ValueError('Negative probability found in TabulatedPDF.',function)

        # Compute the cumulative distribution function = int(pdf(x),x)
        cdf = np.cumsum(pdf)

        # Quietly renormalize the probability if it wasn't already normalized
        totalprobability = cdf[-1]
        cdf /= totalprobability

        self.inverse_cdf_interp = interpolate.interp1d(cdf, xarray, kind='linear')
        self.cdf_interp = interpolate.interp1d(xarray, cdf, kind='linear')
        self.pdf_interp = interpolate.interp1d(xarray, pdf_x, kind='linear')

    def __call__(self, x):
        u"""Return the logarithm of the probability density."""
        if x<self.lower:
            return -np.inf
        elif x>self.upper:
            return -np.inf
        return np.log(self.pdf_interp(x))

    def sample(self, n):
        u"""Use interpolation of inverse CDF to give us `n` random samples from a the distribution."""
        if n is None:
            n = 1 
        return self.inverse_cdf_interp(np.random.rand(n))

    def denormalize_from_prior(self, y):
        u"""Get the value for which the cumulated probability is `y`."""
        return self.inverse_cdf_interp(y)

    def __str__(self):
        u"""Tersely describe ourself to a human mathematician."""
        return "Tabulated transform from {0} on range [{1}, {2}]".format(self.function_filename, self.lower, self.upper)

    def truncate(self, lower, upper):
        u"""Return a new distribution whose range is the intersection of ours with [`lower`, `upper`].

        A :class:`ValueError` will be thrown if the arguments supplied do
        not make sense.

        """
        lower = max(self.lower, lower)
        upper = min(self.upper, upper)
        if lower>upper:
            raise ValueError("One of your priors is inconsistent with the range described in the values file")
        return TabulatedPDF(self.function_filename, lower, upper)


class UniformPrior(Prior):

    u"""Statistical distribution in which all values in a range have equal probability of occurring."""
    
    def __init__(self, a, b):
        u"""Create an object which encapsulates a Uniform distribution over the interval [`a`, `b`]."""
        self.a=a
        self.b=b
        self.norm = -np.log(b-a)
        super(UniformPrior,self).__init__()


    def __call__(self, x):
        u"""Return the logarithm of the probability density, a constant value independent of `x` so long as `x` is in the proper range."""
        if x<self.a or x>self.b:
            return -np.inf
        return self.norm


    def sample(self, n):
        u"""Use NumPy to obtain a random number in the range [`a`, `b`]."""
        return np.random.uniform(self.a, self.b, n)


    def denormalize_from_prior(self, y):
        u"""Interpolate the cumulated probability `y` to the corresponding value in the interval [`a`, `b`]."""
        if y<0.0:
            x = np.nan
        elif y>1.0:
            x = np.nan
        else:
            x = y * (self.b-self.a) + self.a
        return x


    def __str__(self):
        u"""Tersely describe ourself to a human mathematician."""
        return "U({}, {})".format(self.a,self.b)


    def truncate(self, lower, upper):
        u"""Return a new Uniform distribution whose range is the intersection of ours with [`lower`, `upper`].

        A :class:`ValueError` will be thrown if the arguments supplied do
        not make sense.

        """
        a = max(lower, self.a)
        b = min(upper, self.b)
        if a>b:
            raise ValueError("One of your priors is inconsistent with the range described in the values file")
        return UniformPrior(a, b)



class GaussianPrior(Prior):

    u"""Encapsulation of a Normal distribution function."""

    def __init__(self, mu, sigma):
        u"""Make a Normal distribution object with mean `mu` and deviation `sigma`."""
        self.mu = mu
        self.sigma = sigma
        self.sigma2 = sigma**2
        self.norm=0.5*np.log(2*np.pi*self.sigma2)
        super(GaussianPrior,self).__init__()

    def __call__(self, x):
        u"""Return the logarithm of the probability density at `x`."""
        return -0.5 * (x-self.mu)**2 / self.sigma2 - self.norm

    def sample(self, n):
        u"""Use NumPy to give us `n` random samples from a Normal distribution."""
        if n is None:
            n = 1        
        return np.random.normal(self.mu, self.sigma, n)

    def denormalize_from_prior(self, y):
        u"""Obtain the value such that the cumulated probability of obtaining a lesser value is `y`.

        Note that this is a very expensive function to call.

        """
        x_normal = normal_ppf(y)
        x = x_normal*self.sigma + self.mu
        return x

    def __str__(self):
        u"""Tersely describe ourself to a human mathematician."""
        return "N({}, {} ** 2)".format(self.mu, self.sigma)

    def truncate(self, lower, upper):
        u"""Return a :class:`TruncatedGaussianPrior` object, with our mean and variance and the given `lower` and `upper` limits of non-zero probability."""
        return TruncatedGaussianPrior(self.mu, self.sigma, lower, upper)

    
    
class TruncatedGaussianPrior(Prior):

    u"""A Normal distribution, except that probabilities outside of a specified range are truncated to zero."""

    def __init__(self, mu, sigma, lower, upper):
        u"""Get a concrete object representing a distribution with mode at `mu`, ‘shape width’ `sigma`, and limits of non-zero probability `lower` and `upper`."""
        self.lower = lower
        self.upper = upper
        self.mu = mu
        self.sigma = sigma
        self.sigma2 = sigma**2
        self.a = (lower-mu)/sigma
        self.b = (upper-mu)/sigma
        self.phi_a = normal_cdf(self.a)
        self.phi_b = normal_cdf(self.b)
        self.norm = np.log(self.phi_b - self.phi_a) + 0.5*np.log(2*np.pi*self.sigma2)
        super(TruncatedGaussianPrior,self).__init__()

    def __call__(self, x):
        u"""Get the logarithm of the probability density at the value `x`."""
        if x<self.lower:
            return -np.inf
        elif x>self.upper:
            return -np.inf
        return -0.5 * (x-self.mu)**2 / self.sigma2 - self.norm

    def denormalize_from_prior(self, y):
        u"""Get the value for which the cumulated probability is `y`.

        This is a very expensive function.

        """
        x_normal = truncated_normal_ppf(y, self.a, self.b)
        x = x_normal*self.sigma + self.mu
        return x

    def __str__(self):
        u"""Return a terse description of ourself."""
        return "N({}, {} ** 2)   [{} < x < {}]".format(self.mu, self.sigma, self.lower, self.upper)

    def truncate(self, lower, upper):
        u"""Produce a new :class:`Prior` object representing a Normal distribution whose range of non-zero probability is the intersection of our own range with [`lower`, `upper`]."""
        lower = max(self.lower, lower)
        upper = min(self.upper, upper)
        return TruncatedGaussianPrior(self.mu, self.sigma, lower, upper)



class ExponentialPrior(Prior):

    u"""The Exponential Distribution, zero probability of any value less than zero."""

    def __init__(self, beta):
        u"""Create object representing distribution with ‘width’ `beta`."""
        self.beta = beta
        self.log_beta = np.log(beta)
        super(ExponentialPrior,self).__init__()
    
    def __call__(self, x):
        u"""Return logarithm of probability density at `x`."""
        if x<0.0:
            return -np.inf
        return -x/self.beta - self.log_beta
    
    def sample(self,n):
        u"""Use NumPy to obtain random sample of `n` values from Exponential Distribution with our width `beta`."""
        if n is None:
            n = 1        
        return np.random.exponential(self.beta,n)

    def denormalize_from_prior(self, y):
        u"""Return value for which cumulated probability of lesser values occurring is `y`."""
        #y = 1 - exp(-x/beta)
        #exp(-x/beta) = 1 - y
        #-x/beta = log(1-y)
        #x = -beta * log(1-y)
        return -self.beta * np.log(1-y)
        
    def __str__(self):
        u"""Give a terse description of ourself."""
        return "Expon({})".format(self.beta)

    def truncate(self, lower, upper):
        u"""Return a :class:`Prior` object representing the distribution you get when you take the current distribution but set probability to zero everywhere outside the range [`lower`, `upper`], and re-normalize."""
        return TruncatedExponentialPrior(self.beta, lower, upper)



class TruncatedExponentialPrior(Prior):

    u"""Like the Exponential prior, but truncated."""

    def __init__(self, beta, lower, upper):
        u"""Create a distribution with ‘half-life’ `beta`, `lower` bound of non-zero probability, and `upper` bound."""
        self.beta = beta
        self.log_beta = np.log(beta)
        if lower<0:
            lower = 0.0
        self.lower = lower
        self.upper = upper
        self.a = lower/beta
        self.b = upper/beta
        self.phi_a = exponential_cdf(self.a)
        self.phi_b = exponential_cdf(self.b)
        self.norm = np.log(self.phi_b - self.phi_a) + self.log_beta
        super(TruncatedExponentialPrior,self).__init__()

    
    def __call__(self, x):
        u"""Return the logarithm of probability density at `x`."""
        #(1/beta)*exp(-x/beta)
        if x<self.lower:
            return -np.inf
        if x>self.upper:
            return -np.inf
        return -x/self.beta - self.norm
    

    def denormalize_from_prior(self, y):
        u"""Return the value at which the cumulated probability is `y`."""
        x_normal = truncated_exponential_ppf(y, self.a, self.b)
        x = x_normal * self.beta
        return x

    def __str__(self):
        u"""Give a terse description of ourself."""
        return "Expon({})   [{} < x < {}]".format(self.beta, self.lower, self.upper)

    def truncate(self, lower, upper):
        u"""Return a :class:`Prior` like ourself but with range of non-zero probability further restricted to `lower` and `upper` bounds."""
        lower = max(lower, self.lower)
        upper = min(upper, self.upper)
        return TruncatedExponentialPrior(self.beta, lower, upper)


class TruncatedOneoverxPrior(Prior):
    u"""The 1/x distribution, which is a uniform distribution in ln(x). As ln(x) diverges in both directions, we only provide the truncated option."""

    def __init__(self, lower, upper):
        u"""Create a distribution with 1/x, `lower` bound of non-zero probability, and `upper` bound."""
        if lower<=0:
            lower = np.nextafter(0, 1)
        self.lower = lower
        self.upper = upper
        self.ln_lower = np.log(lower)
        self.ln_upper = np.log(upper)
        # Normalization: \int_upper^lower 1/x dx = ln(upper) - ln(lower)
        self.norm = self.ln_upper-self.ln_lower
        self.ln_norm = np.log(self.norm)
        super(TruncatedOneoverxPrior,self).__init__()

    def __call__(self, x):
        u"""Return the logarithm of probability density at `x`."""
        if x<self.lower:
            return -np.inf
        if x>self.upper:
            return -np.inf
        return -np.log(x) - self.ln_norm

    def sample(self, n):
        u"""Obtain random sample of `n` values from 1/x distribution within `lower` and `upper` bounds."""
        if n is None:
            n = 1
        return np.exp(np.random.uniform(self.ln_lower, self.ln_upper, n))

    def denormalize_from_prior(self, y):
        u"""Return the value at which the cumulated probability is `y`."""
        # int_a^x dz 1/z 1/N = (ln(x)-ln(a)) / N != y
        # ln(x) = y*N + ln(a)
        return np.exp(y*self.norm + self.ln_lower)

    def __str__(self):
        u"""Give a terse description of ourself."""
        return "1/x   [{} < x < {}]".format(self.lower, self.upper)

    def truncate(self, lower, upper):
        u"""Return a :class:`Prior` like ourself but with range of non-zero probability further restricted to `lower` and `upper` bounds."""
        lower = max(lower, self.lower)
        upper = min(upper, self.upper)
        return TruncatedOneoverxPrior(lower, upper)



class DeltaFunctionPrior(Prior):

    u"""Probability distribution with non-zero probability at a single value."""

    # In case this is useful later on
    def __init__(self, x0):
        u"""Create object with atom of probability at `x0`."""
        self.x0 = x0
        super(DeltaFunctionPrior,self).__init__()

    def __call__(self, x):
        u"""The log-density is zero when `x` is `x0`, minus infinity otherwise."""
        if x==self.x0:
            return 0.0
        return -np.inf

    def sample(self, n=None):
        u"""Just return `x0` `n` times."""
        if n is None:
            n = 1
        return np.repeat(self.x0, n)

    def __str__(self):
        u"""Terse description of ourself."""
        return "delta({})".format(self.x0)

    def denormalize_from_prior(self, x):
        u"""Just return `x0`; itʼs the only value with any probability."""
        return self.x0
        


# Helper functions
def inverse_function(f, y, xmin, xmax, *args, **kwargs):
    "Find x in [xmin,xmax] such that f(x)==y, in 1D, with bisection"
    import scipy.optimize
    def g(x):
        return f(x, *args, **kwargs) - y
    x = scipy.optimize.bisect(g, xmin, xmax)
    return x

SQRT2 = np.sqrt(2.)
def normal_cdf(x):
#    return 0.5*math.erf(x) + 0.5
    return 0.5*(math.erf(x/SQRT2) + 1)


def normal_ppf(y):
    if y<0:
        return np.nan
    if y>1:
        return np.nan
    return inverse_function(normal_cdf, y, -20.0, 20.0)

def truncated_normal_cdf(x, a, b):
    if x<a:
        return np.nan
    if x>b:
        return np.nan
    phi_a = normal_cdf(a)
    phi_b = normal_cdf(b)
    phi_x = normal_cdf(x)
    return (phi_x - phi_a) / (phi_b - phi_a)

def truncated_normal_ppf(y, a, b):
    if y<0:
        return np.nan
    if y>1:
        return np.nan
    return inverse_function(truncated_normal_cdf, y, a, b, a, b)

def exponential_cdf(x):
    if x<0.0:
        return np.nan
    return 1 - np.exp(-x)

def truncated_exponential_cdf(x, a, b):
    if x<a:
        return np.nan
    if x>b:
        return np.nan
    phi_a = exponential_cdf(a)
    phi_b = exponential_cdf(b)
    phi_x = exponential_cdf(x)
    return (phi_x - phi_a) / (phi_b - phi_a)

def exponential_ppf(y):
    #y = 1 - exp(-x)
    # exp(-x) = 1-y
    # x = -log(1-y)
    return -np.log(1-y)

def truncated_exponential_ppf(y, a, b):
    if y<0:
        return np.nan
    if y>1:
        return np.nan    
    return inverse_function(truncated_exponential_cdf, y, a, b, a, b)


"""
Stop!
This file is not ready yet.

I think I need to re-write the plotting code almost from scratch
as it's grown pretty organically.

"""

import matplotlib
matplotlib.use('Agg')
#matplotlib.rcParams['text.usetex']=True
matplotlib.rcParams['font.family']='serif'
matplotlib.rcParams['font.size']=18
matplotlib.rcParams['legend.fontsize']=15
matplotlib.rcParams['xtick.major.size'] = 10.0
matplotlib.rcParams['xtick.minor.size'] = 5.0
matplotlib.rcParams['ytick.major.size'] = 10.0
matplotlib.rcParams['ytick.minor.size'] = 5.0
import pylab
import numpy as np
import collections
from .plotter import Plotter
import itertools
import scipy.optimize
try:
	from cosmosis import output as output_module
except ImportError:
	print("Running without cosmosis: no pretty section names or running on ini files")


class GridPlotter(Plotter):
	#def __init__(self, *args, **kwargs):
		# super(GridPlotter, self).__init__(*args, **kwargs)
		# convert the loaded chain data sets into grids
		# need the number of varied parameters
		# how do we get that?

	@classmethod
	def from_outputs(cls, options, **kw):
		column_names, chains, metadata, comments, final_metadata = output_module.input_from_options(options)
		chains = np.vstack(chains).T
		chains = dict(list(zip(column_names,chains)))
		chain_data = {"Chains":chains}
		return cls(chain_data, **kw)

	@classmethod
	def from_chain_files(cls, filenames, **kw):
		nameset = [open(filename).readline().strip().strip('#').replace("/","--").split() for filename in filenames]
		dataset = [(np.loadtxt(filename).T) for filename in filenames]
		chain_data = collections.OrderedDict()
		for filename,names,data in zip(filenames,nameset,dataset):
			chain_datum = dict(list(zip(names,data)))
			chain_data[filename] = chain_datum	
		return cls(chain_data, **kw)

	def _plot_1d(self, name1):
		cols1 = list(self.cols_for_name(name1).values())[0]
		like = list(self.cols_for_name("LIKE").values())[0]
		vals1 = np.unique(cols1)
		n1 = len(vals1)

		#check for derived parameters - this is a weak check
		if len(vals1)==len(cols1) and not len(self.all_names)==2:
			print("Not grid-plotting %s as it seems to be a derived parameter" % name1)
			raise ValueError(name1)
		like_sum = np.zeros(n1)

		#marginalize
		for k,v1 in enumerate(vals1):
			w = np.where(cols1==v1)
			like_sum[k] = np.log(np.exp(like[w]).sum())
		like = like_sum.flatten()

		#linearly interpolate
		n1 *= 10
		vals1_interp = np.linspace(vals1[0], vals1[-1], n1)
		like_interp = np.interp(vals1_interp, vals1, like)
		vals1 = vals1_interp
		like = like_interp

		#normalize
		like -= like.max()

		#Determine the spacing in the different parameters
		dx = vals1[1]-vals1[0]

		ax = pylab.gca()
		pylab.xlim(cols1.min()-dx/2., cols1.max()+dx/2.)
		pylab.ylim(0,1.05)
		pylab.plot(vals1, np.exp(like), linewidth=3)

		#Find the levels of the 68% and 95% contours
		level1, level2 = self._find_grid_contours(like, 0.68, 0.95)
		above = np.where(like>level1)[0]
		left = above[0]
		right = above[-1]
		x1 = vals1[left]
		x2 = vals1[right]
		#Lmin = ax.get_ylim()[0]
		L1 = like[left]
		L2 = like[right]
		pylab.plot([x1,x1], [0, np.exp(L1)], ':', color='black')
		pylab.plot([x2,x2], [0, np.exp(L2)], ':', color='black')


		above = np.where(like>level2)[0]
		left = above[0]
		right = above[-1]
		x1 = vals1[left]
		x2 = vals1[right]
		#Lmin = ax.get_ylim()[0]
		L1 = like[left]
		L2 = like[right]
		pylab.plot([x1,x1], [0, np.exp(L1)], ':', color='gray')
		pylab.plot([x2,x2], [0, np.exp(L2)], ':', color='gray')
		pylab.xlabel("$"+self._display_names[name1]+"$")


	@staticmethod
	def _find_grid_contours(log_like, contour1, contour2):
		like = np.exp(log_like)
		like_total = like.sum()
		def objective(limit, target):
			w = np.where(like>limit)
			return like[w].sum() - target
		target1 = like_total*contour1
		target2 = like_total*contour2
		level1 = scipy.optimize.bisect(objective, like.min(), like.max(), args=(target1,))
		level2 = scipy.optimize.bisect(objective, like.min(), like.max(), args=(target2,))
		level1 = np.log(level1)
		level2 = np.log(level2)
		return level1, level2


	def _plot_2d(self, name1, name2, log_like=True):
		# Load the columns
		cols1 = list(self.cols_for_name(name1).values())[0]
		cols2 = list(self.cols_for_name(name2).values())[0]
		like = list(self.cols_for_name("LIKE").values())[0]

		#Marginalize over all the other parameters by summing
		#them up
		vals1 = np.unique(cols1)
		vals2 = np.unique(cols2)
		n1 = len(vals1)
		n2 = len(vals2)
		like_sum = np.zeros((n1,n2))
		for k,(v1, v2) in enumerate(itertools.product(vals1, vals2)):
			w = np.where((cols1==v1)&(cols2==v2))
			i,j = np.unravel_index(k, like_sum.shape)
			like_sum[i,j] = np.log(np.exp(like[w]).sum())
		like = like_sum.flatten()

		#Normalize the log-likelihood to peak=0
		like -= like.max()

		#Determine the spacing in the different parameters
		dx = vals1[1]-vals1[0]
		dy = vals2[1]-vals2[0]

		# Set up the axis ranges, grid, and labels
		ax = pylab.gca()
		pylab.xlim(cols1.min()-dx/2., cols1.max()+dx/2.)
		pylab.ylim(cols2.min()-dy/2., cols2.max()+dy/2.)
		# pylab.grid()
		pylab.xlabel("$"+self._display_names[name1]+"$")
		pylab.ylabel("$"+self._display_names[name2]+"$")

		#Choose a color mapping
		norm = matplotlib.colors.Normalize(np.exp(like.min()), np.exp(like.max()))
		colormap = pylab.cm.Reds


		edges1 = set()
		edges2 = set()
		def toggle_edge(edges, x1,y1,x2,y2):
			e = ((np.around(x1,4),np.around(y1,4)), (np.around(x2,4),np.around(y2,4)))
			if e in edges: edges.remove(e)
			else: edges.add(e)

		#Find the levels of the 68% and 95% contours
		level1, level2 = self._find_grid_contours(like, 0.68, 0.95)


		#Loop through our grid
		for (px, py, L) in zip(cols1, cols2,like):
			#get the color for the point
			c = colormap(norm(np.exp(L)))
			#create and apply the square colour patch
			r = pylab.Rectangle((px-dx/2.,py-dy/2.), dx, dy, color=c)
			ax.add_artist(r)
			if L>level1:
				toggle_edge(edges1, px-dx/2., py-dy/2., px-dx/2., py+dy/2.)
				toggle_edge(edges1, px-dx/2., py-dy/2., px+dx/2., py-dy/2.)
				toggle_edge(edges1, px-dx/2., py+dy/2., px+dx/2., py+dy/2.)
				toggle_edge(edges1, px+dx/2., py-dy/2., px+dx/2., py+dy/2.)
			if L>level2:
				toggle_edge(edges2, px-dx/2., py-dy/2., px-dx/2., py+dy/2.)
				toggle_edge(edges2, px-dx/2., py-dy/2., px+dx/2., py-dy/2.)
				toggle_edge(edges2, px-dx/2., py+dy/2., px+dx/2., py+dy/2.)
				toggle_edge(edges2, px+dx/2., py-dy/2., px+dx/2., py+dy/2.)

		for ((x1,y1), (x2,y2)) in edges1:
			pylab.plot([x1,x2],[y1,y2], '-', linewidth=3, color='black')
		for ((x1,y1), (x2,y2)) in edges2:
			pylab.plot([x1,x2],[y1,y2], '-', linewidth=3, color='gray')

		#create and add a colorbar
		sm = pylab.cm.ScalarMappable(cmap=colormap, norm=norm)
		sm._A = [] #hack from StackOverflow to make this work
		pylab.colorbar(sm, label='Likelihood')




from cosmosis.runtime import Inifile, register_new_parameter, LikelihoodPipeline, Parameter, Module
from cosmosis.datablock import DataBlock
from cosmosis.samplers.sampler import Sampler
from cosmosis.runtime.prior import TruncatedGaussianPrior, DeltaFunctionPrior
from cosmosis.output.in_memory_output import InMemoryOutput
import numpy as np
import os
import tempfile

root = os.path.split(os.path.abspath(__file__))[0]

def test_add_param():

    sampler_class = Sampler.registry["emcee"]

    values = tempfile.NamedTemporaryFile('w')
    values.write(
        "[parameters]\n"
        "p1=-3.0  0.0  3.0\n"
        "p2=-3.0  0.0  3.0\n")
    values.flush()

    override = {
        ('runtime', 'root'): root,
        ("pipeline", "debug"): "F",
        ("pipeline", "quiet"): "T",
        ("pipeline", "modules"): "test2",
        ("pipeline", "values"): values.name,
        ("test2", "file"): "test_module2.py",
        ("emcee", "walkers"): "8",
        ("emcee", "samples"): "10"
    }

    ini = Inifile(None, override=override)

    # Make the pipeline itself
    pipeline = LikelihoodPipeline(ini)

    # test the the new added parameter has worked
    assert len(pipeline.varied_params) == 3
    p = pipeline.varied_params[2]
    assert str(p) == "new_parameters--p3"
    assert isinstance(p, Parameter)
    assert isinstance(p.prior, TruncatedGaussianPrior)
    assert np.isclose(p.prior.mu, 0.1)
    assert np.isclose(p.prior.sigma, 0.2)

    assert len(pipeline.fixed_params) == 1
    p = pipeline.fixed_params[0]
    assert isinstance(p, Parameter)
    assert isinstance(p.prior, DeltaFunctionPrior)


    output = InMemoryOutput()
    sampler = sampler_class(ini, pipeline, output)
    sampler.config()


    # check that the output is working
    assert output.column_index_for_name("new_parameters--p3") == 2

    while not sampler.is_converged():
        sampler.execute()


    p1 = output['parameters--p1']
    p2 = output['parameters--p2']
    p3 = output['new_parameters--p3']
    assert p3.max() < 1.0
    assert p3.min() > -1.0

    return output

def test_missing_setup():
    # check the register_new_parameter feature when no
    # setup is currently happening
    module = Module("test2", root + "/test_module2.py")
    config = DataBlock()
    module.setup(config)

def test_unused_param_warning(capsys):
    # check that an appropriate warning is generated
    # when a parameter is unused
    module = Module("test", root + "/test_module.py")
    config = DataBlock()
    config['test', 'unused'] = "unused_parameter"
    module.setup(config)
    out, _ = capsys.readouterr()
    assert "**** WARNING: Parameter 'unused'" in out



if __name__ == '__main__':
    test_add_param()
    test_missing_setup()



from .. import Sampler
import numpy as np


class MaxlikeSampler(Sampler):
    sampler_outputs = [("prior", float), ("like", float), ("post", float)]

    def config(self):
        self.tolerance = self.read_ini("tolerance", float, 1e-3)
        self.maxiter = self.read_ini("maxiter", int, 1000)
        self.output_ini = self.read_ini("output_ini", str, "")
        self.output_cov = self.read_ini("output_covmat", str, "")
        self.method = self.read_ini("method",str,"Nelder-Mead")
        self.max_posterior = self.read_ini("max_posterior", bool, False)

        if self.max_posterior:
            print("------------------------------------------------")
            print("NOTE: Running optimizer in **max-posterior** mode:")
            print("NOTE: Will maximize the combined likelihood and prior")
            print("------------------------------------------------")
        else:
            print("--------------------------------------------------")
            print("NOTE: Running optimizer in **max-like** mode:")
            print("NOTE: not including the prior, just the likelihood.")
            print("NOTE: Set the parameter max_posterior=T to change this.")
            print("NOTE: This won't matter unless you set some non-flat")
            print("NOTE: priors in a separate priors file.")
            print("--------------------------------------------------")

        self.converged = False

    def execute(self):
        import scipy.optimize

        def likefn(p_in):
            #Check the normalization
            if (not np.all(p_in>=0)) or (not np.all(p_in<=1)):
                return np.inf
            p = self.pipeline.denormalize_vector(p_in)
            if self.max_posterior:
                like, extra = self.pipeline.posterior(p)
                self.output.log_debug("%s  post=%f"%('   '.join(str(x) for x in p),like))
            else:
                like, extra = self.pipeline.likelihood(p)
                self.output.log_debug("%s  like=%f"%('   '.join(str(x) for x in p),like))
            return -like

        # starting position in the normalized space.  This will be taken from
        # a previous sampler if available, or the values file if not.
        start_vector = self.pipeline.normalize_vector(self.start_estimate())
        bounds = [(0.0, 1.0) for p in self.pipeline.varied_params]

        # check that the starting position is a valid point
        start_like = likefn(start_vector)
        if not np.isfinite(start_like):
            raise RuntimeError('invalid starting point for maxlike')

        result = scipy.optimize.minimize(likefn, start_vector, method=self.method, 
          jac=False, tol=self.tolerance,  #bounds=bounds, 
          options={'maxiter':self.maxiter, 'disp':True})

        opt_norm = result.x
        opt = self.pipeline.denormalize_vector(opt_norm)
        

        #Some output - first log the parameters to the screen.
        #It's not really a warning - that's just a level name
        results = self.pipeline.run_results(opt)
        if self.max_posterior:

            self.output.log_warning("Best fit (by posterior):\n%s"%'   '.join(str(x) for x in opt))
        else:
            self.output.log_warning("Best fit (by likelihood):\n%s"%'   '.join(str(x) for x in opt))
        self.output.log_warning("Posterior: {}\n".format(results.post))
        self.output.log_warning("Likelihood: {}\n".format(results.like))

        #Next save them to the proper table file
        self.output.parameters(opt, results.extra, results.prior, results.like, results.post)

        #If requested, create a new ini file for the
        #best fit.
        if self.output_ini:
          self.pipeline.create_ini(opt, self.output_ini)

        self.distribution_hints.set_peak(opt)          

        #Also if requested, approximate the covariance matrix with the 
        #inverse of the Hessian matrix.
        #For a gaussian likelihood this is exact.
        covmat = None
        if hasattr(result, 'hess_inv'):
            if self.method == "L-BFGS-B":
                covmat = self.pipeline.denormalize_matrix(result.hess_inv.todense())
            else:
                covmat = self.pipeline.denormalize_matrix(result.hess_inv)
        elif hasattr(result, 'hess'):
            covmat = self.pipeline.denormalize_matrix(np.linalg.inv(result.hess_inv))

        if covmat is None:
            if self.output_cov:
               self.output.log_error("Sorry - the optimization method you chose does not return a covariance (or Hessian) matrix")
        else:
            if self.output_cov:
                np.savetxt(self.output_cov, covmat)
            self.distribution_hints.set_cov(covmat)

        self.converged = True

    def is_converged(self):
        return self.converged



#!/usr/bin/env python


import sys
import configparser

import argparse
import os
import pdb
from .runtime.config import Inifile, CosmosisConfigurationError
from .runtime.pipeline import LikelihoodPipeline
from .runtime import mpi_pool
from .runtime import process_pool
from .runtime.utils import ParseExtraParameters, stdout_redirected, import_by_path
from .samplers.sampler import Sampler, ParallelSampler, Hints
from . import output as output_module
from .runtime.handler import activate_segfault_handling


RUNTIME_INI_SECTION = "runtime"


def demo_1_special (args):
    if "demo1.ini" in args.inifile:
        print()
        print("Congratulations: you have just run cosmosis demo one!")
        if os.path.exists("./conda"):
            print() 
            print("You can make plots of the outputs of this using this command:")
            print("  postprocess demos/demo1.ini -o plots -p demo1")
            print()
            print("If you get a message about 'Abort Trap 6' then see the FAQ:")
            print("https://bitbucket.org/joezuntz/cosmosis/wiki/FAQ")
            print()
            print("Then you can try out the other demos...")
            print("... and read the information about plotting their output and what they are doing online.")
            print("Please get in touch with any problems, ideally by filing an Issue. Thanks!")
        else:
            print("You can make plots of the outputs of this using the command:")
            print()
            print("postprocess demos/demo1.ini -o plots -p demo1")
            print()
            print("Then you can try out the other demos...")
            print("... and read the information about plotting their output and what they are doing online.")
            print("Please get in touch with any problems, ideally by filing an Issue. Thanks!")
            print()


def demo_10_special (args):
    if   "demo10.ini" in args.inifile   and   not os.getenv ("HALOFIT", ""):
        print()
        print("Welcome to demo10!")
        print()
        print("**PLEASE NOTE**:")
        print()
        print("There are two flavours of this demo, selected through an ")
        print("environment variable called `HALOFIT'; this variable is not ")
        print("currently set, so we are giving it the value `halofit'.")
        print("Please see the wiki for more information: ")
        print("https://bitbucket.org/joezuntz/cosmosis/wiki/Demo10.")

        os.environ ["HALOFIT"] = "halofit"



def demo_20a_special (args):
    if  "demo20a.ini" in args.inifile:
        print ()
        print ("You have completed demo20a, now run demo20b and compare")
        print ("results with demo5!")


def demo_20b_special (args):
    if   "demo20b.ini" in args.inifile   and   not os.path.isfile ("./demo20a.txt"):
        print ()
        print ("********************************************************")
        print ("*** YOU MUST RUN demo20a BEFORE YOU CAN RUN demo20b. ***")
        print ("********************************************************")

def sampler_main_loop(sampler, output, pool, is_root):
    # Run the sampler until convergence
    # which really means "finished" here - 
    # a sampler can "converge" just by reaching the 
    # limit of the number of samples it is allowed.
    if is_root:
        while not sampler.is_converged():
            sampler.execute()
            #Flush any output. This is to stop
            #a problem in some MPI cases where loads
            #of output is built up before being written.
            if output:
                output.flush()
        # If we are in parallel tell the other processors to end the 
        # loop and prepare for the next sampler
        if pool and sampler.is_parallel_sampler:
            pool.close()
    else:
        if sampler.is_parallel_sampler:
            sampler.worker()



def write_header_output(output, params, values, pipeline):
    # If there is an output file, save the ini information to
    # it as well.  We do it here since it's nicer to have it
    # after the sampler options that get written in sampler.config.
    # Create a buffer to store the output:
    output.comment("START_OF_PARAMS_INI")
    comment_wrapper = output.comment_file_wrapper()
    params.write(comment_wrapper)
    output.comment("END_OF_PARAMS_INI")
    # Do the same with the values file.
    # Unfortunately that means reading it in again;
    # if we ever refactor this bit we could eliminate that.
    if isinstance(values, Inifile):
        values_ini = values
    elif values is None:
        values_ini=Inifile(pipeline.values_filename)
    else:
        values_ini=values
    output.comment("START_OF_VALUES_INI")
    values_ini.write(comment_wrapper)
    output.comment("END_OF_VALUES_INI")

    # And the same with the priors
    output.comment("START_OF_PRIORS_INI")
    for priors_file in pipeline.priors_files:
        if isinstance(priors_file, Inifile):
            prior_ini = priors_file
        else:
            prior_ini=Inifile(priors_file)
        prior_ini.write(comment_wrapper)
    output.comment("END_OF_PRIORS_INI")

def setup_output(sampler_class, sampler_number, ini, pool, number_samplers, sample_method, resume):

    needs_output = sampler_class.needs_output and \
       (pool is None or pool.is_master() or sampler_class.parallel_output)

    if not needs_output:
        return None


    #create the output files and methods.
    try:
        output_options = dict(ini.items('output'))
    except configparser.NoSectionError:
        raise ValueError("ERROR:\nFor the sampler (%s) you chose in the [runtime] section of the ini file I also need an [output] section describing how to save results\n\n"%sample_method)
    #Additionally we tell the output here if
    #we are parallel or not.
    if (pool is not None) and (sampler_class.parallel_output):
        output_options['rank'] = pool.rank
        output_options['parallel'] = pool.size

    #Give different output filenames to the different sampling steps
    #Only change if this is not the last sampling step - the final
    #one retains the name in the output file.
    # Change, e.g. demo17.txt to demo17.fisher.txt
    if ("filename" in output_options) and (sampler_number<number_samplers-1):
        filename = output_options['filename']
        filename, ext = os.path.splitext(filename)
        filename += '.' + sampler_name
        filename += ext
        output_options['filename'] = filename


    #Generate the output from a factory
    output = output_module.output_from_options(output_options, resume)
    output.metadata("sampler", sample_method)

    if ("filename" in output_options):
        print("* Saving output -> {}".format(output_options['filename']))

    return output


def run_cosmosis(args, pool=None, ini=None, pipeline=None, values=None):
    # In case we need to hand-hold a naive demo-10 user.

    # Load configuration.
    is_root = (pool is None) or pool.is_master()
    if ini is None:
        ini = Inifile(args.inifile, override=args.params, print_include_messages=is_root)

    pre_script = ini.get(RUNTIME_INI_SECTION, "pre_script", fallback="")
    post_script = ini.get(RUNTIME_INI_SECTION, "post_script", fallback="")

    if is_root:
        # This decodes the exist status
        status = os.WEXITSTATUS(os.system(pre_script))
        if status:
            raise RuntimeError("The pre-run script {} retuned non-zero status {}".format(
                pre_script, status))

    if is_root and args.mem:
        from cosmosis.runtime.memmon import MemoryMonitor
        # This launches a memory monitor that prints out (from a new thread)
        # the memory usage every args.mem seconds
        mem = MemoryMonitor.start_in_thread(interval=args.mem)

    # Create pipeline.
    if pipeline is None:
        cleanup_pipeline = True
        pool_stdout = ini.getboolean(RUNTIME_INI_SECTION, "pool_stdout", fallback=False)
        if is_root or pool_stdout:
            pipeline = LikelihoodPipeline(ini, override=args.variables, values=values, only=args.only)
        else:
            # Suppress output on everything except the master process
            if pool_stdout:
                pipeline = LikelihoodPipeline(ini, override=args.variables, only=args.only) 
            else:
                with stdout_redirected():
                    pipeline = LikelihoodPipeline(ini, override=args.variables, only=args.only) 

        if pipeline.do_fast_slow:
            pipeline.setup_fast_subspaces()
    else:
        # We should not cleanup a pipeline which we didn't make
        cleanup_pipeline = False

    # This feature lets us import additional samplers at runtime
    sampler_files = ini.get(RUNTIME_INI_SECTION, "import_samplers", fallback="").split()
    for i, sampler_file in enumerate(sampler_files):
        # give the module a new name to avoid name clashes if people
        # just call their thing by the same name
        import_by_path('additional_samplers_{}'.format(i), sampler_file)



    # determine the type(s) of sampling we want.
    sample_methods = ini.get(RUNTIME_INI_SECTION, "sampler", fallback="test").split()

    for sample_method in sample_methods:
        if sample_method not in Sampler.registry:
            raise ValueError("Unknown sampler method %s" % (sample_method,))

    #Get that sampler from the system.
    sampler_classes = [Sampler.registry[sample_method] for sample_method in sample_methods]

    if pool:
        if not any(issubclass(sampler_class,ParallelSampler) for sampler_class in sampler_classes):
            if len(sampler_classes)>1:
                raise ValueError("None of the samplers you chose support parallel execution!")
            else:
                raise ValueError("The sampler you chose does not support parallel execution!")
        for sampler_class in sampler_classes:
            if isinstance(pool, process_pool.Pool) and issubclass(sampler_class,ParallelSampler) and not sampler_class.supports_smp:
                name = sampler_class.__name__[:-len("Sampler")].lower()
                raise ValueError("Sorry, the {} sampler does not support the --smp flag.".format(name))

    number_samplers = len(sampler_classes)


    #To start with we do not have any estimates of 
    #anything the samplers might give us like centers
    #or covariances. 
    distribution_hints = Hints()

    #Now that we have a sampler we know whether we will need an
    #output file or not.  By default new samplers do need one.
    for sampler_number, (sampler_class, sample_method) in enumerate(
            zip(sampler_classes, sample_methods)):
        sampler_name = sampler_class.__name__[:-len("Sampler")].lower()

        # The resume feature lets us restart from an existing file.
        # It's not fully rolled out to all the suitable samplers yet though.
        resume = ini.getboolean(RUNTIME_INI_SECTION, "resume", fallback=False)

        # Not all samplers can be resumed.
        if resume and not sampler_class.supports_resume:
            print("NOTE: You set resume=T in the [runtime] section but the sampler {} does not support resuming yet.  I will ignore this option.".format(sampler_name))
            resume=False

        if is_root:
            print("****************************")
            print("* Running sampler {}/{}: {}".format(sampler_number+1,number_samplers, sampler_name))

        output = setup_output(sampler_class, sampler_number, ini, pool, number_samplers, sample_method, resume)
        print("****************************")

        #Initialize our sampler, with the class we got above.
        #It needs an extra pool argument if it is a ParallelSampler.
        #All the parallel samplers can also act serially too.
        if pool and sampler_class.is_parallel_sampler:
            sampler = sampler_class(ini, pipeline, output, pool)
        else:
            sampler = sampler_class(ini, pipeline, output)
         
        #Set up the sampler - for example loading
        #any resources it needs or checking the ini file
        #for additional parameters.
        sampler.distribution_hints.update(distribution_hints)
        sampler.config()

        # Potentially resume
        if resume and sampler_class.needs_output and \
            sampler_class.supports_resume and \
            (is_root or sampler_class.parallel_output):
           sampler.resume()

        if output:
            write_header_output(output, ini, values, pipeline)


        sampler_main_loop(sampler, output, pool, is_root)
        distribution_hints.update(sampler.distribution_hints)

        if output:
            output.close()

    if cleanup_pipeline:
        pipeline.cleanup()


    if is_root and args.mem:
        mem.stop()

    # Extra-special actions we take to mollycoddle a brand-new user!
    demo_1_special (args)
    demo_20a_special (args)

    # User can specify in the runtime section a post-run script to launch.
    # In general this may be less useful than the pre-run script, because
    # often chains time-out instead of actually completing.
    # But we still offer it
    if post_script and is_root:
        # This decodes the exist status
        status = os.WEXITSTATUS(os.system(post_script))
        if status:
            sys.stdout.write("WARNING: The post-run script {} failed with error {}".format(
                post_script, error))

    return 0


def make_graph(inifile, dotfile, params=None, variables=None):
    """
    Make a graphviz "dot" format file, describing the pipeline
    and how data is passed from section to section.

    Requires pygraphviz.

    Parameters
    ----------

    inifile: str
        A path to a pipeline file or an Inifile object

    dotfile: str
        Path to the new graph output file

    params: dict or None
        Dictionary of parameter overrides

    variables: dict or None
        Dictionary of value overrides
    """
    ini = Inifile(inifile, override=params)
    pipeline = LikelihoodPipeline(ini, override=variables)
    data = pipeline.run_parameters(pipeline.start_vector())
    pipeline.make_graph(data, dotfile)

def main():
    try:
        parser = argparse.ArgumentParser(description="Run a pipeline with a single set of parameters", add_help=True)
        parser.add_argument("inifile", help="Input ini file of parameters")
        parser.add_argument("--mpi",action='store_true',help="Run in MPI mode.")
        parser.add_argument("--smp",type=int,default=0,help="Run with the given number of processes in shared memory multiprocessing (this is experimental and does not work for multinest).")
        parser.add_argument("--pdb",action='store_true',help="Start the python debugger on an uncaught error. Only in serial mode.")
        parser.add_argument("--segfaults", "--experimental-fault-handling", action='store_true',help="Activate a mode that gives more info on segfault")
        parser.add_argument("--mem", type=int, default=0, help="Print out memory usage every this many seconds from root process")
        parser.add_argument("-p", "--params", nargs="*", action=ParseExtraParameters, help="Override parameters in inifile, with format section.name1=value1 section.name2=value2...")
        parser.add_argument("-v", "--variables", nargs="*", action=ParseExtraParameters, help="Override variables in values file, with format section.name1=value1 section.name2=value2...")
        parser.add_argument("--only", nargs="*", help="Fix all parameters except the ones listed")
        parser.add_argument("--graph", type=str, default='', help="Do not run a sampler; instead make a graphviz dot file of the pipeline")
        args = parser.parse_args(sys.argv[1:])

        if args.graph:
            make_graph(args.inifile, args.graph, args.params, args.variables)
            return

        demo_10_special (args)
        demo_20b_special (args)

        if args.segfaults:
            activate_segfault_handling()

        # initialize parallel workers
        if args.mpi:
            with mpi_pool.MPIPool() as pool:
                return run_cosmosis(args,pool)
        elif args.smp:
            with process_pool.Pool(args.smp) as pool:
                return run_cosmosis(args,pool)
        else:
            try:
                return run_cosmosis(args)
            except Exception as error:
                if args.pdb:
                    print("There was an exception - starting python debugger because you ran with --pdb")
                    print(error)
                    pdb.post_mortem()
                else:
                    raise
    except CosmosisConfigurationError as e:
        print(e)
        return 1

    # Extra-special actions we take to mollycoddle a brand-new user!
    demo_1_special (args)
    demo_20a_special (args)


if __name__=="__main__":
    status = main()
    sys.exit(status)



from .. import ParallelSampler
import numpy as np
from .snake import Snake

def posterior(p_in):
    #Check the normalization
    if (not np.all(p_in>=0)) or (not np.all(p_in<=1)):
        print(p_in)
        return -np.inf, ([np.nan for i in range(len(snake_pipeline.extra_saves))], -np.inf)
    p = snake_pipeline.denormalize_vector(p_in)
    results = snake_pipeline.run_results(p)
    return results.post, (results.extra, results.prior)


class SnakeSampler(ParallelSampler):
    sampler_outputs = [("prior", float), ("post", float)]
    parallel_output = False

    def config(self):
        global snake_pipeline
        snake_pipeline=self.pipeline
        if self.is_master():
            threshold = self.read_ini("threshold", float, 4.0)
            self.grid_size = 1.0/self.read_ini("nsample_dimension", int, 10)
            self.maxiter = self.read_ini("maxiter", int, 100000)

            
            origin = self.pipeline.normalize_vector(self.pipeline.start_vector())
            spacing = np.repeat(self.grid_size, len(self.pipeline.varied_params))
            self.snake = Snake(posterior, origin, spacing, threshold, pool=self.pool)

    def execute(self):
        X, P, E = self.snake.iterate()
        for (x,post,(extra, prior)) in zip(X,P,E):
            try:
                x = self.pipeline.denormalize_vector(x)
                self.output.parameters(x, extra, prior, post)
            except ValueError:
                print("The snake is trying to escape its bounds!")



    def is_converged(self):
        if self.snake.converged():
            print("Snake has converged!")
            print("Best post = %f    Best surface point = %f" %(self.snake.best_like_ever, self.snake.best_fit_like))
            return True
        if self.snake.iterations > self.maxiter:
            print("Run out of iterations.")
            print("Done %d, max allowed %d" % (self.snake.iterations, self.maxiter))
            return True
        return False



import numpy as np
from .proposal.standard import Proposal, FastSlowProposal
from copy import copy


class Bad(object):
    def __init__(self):
        self.post = -np.inf


class MCMC(object):
    def __init__(self, start, posterior, covariance, quiet=False,
        tuning_frequency=-1, tuning_grace=np.inf, tuning_end=np.inf, 
        scaling=2.4,
        exponential_probability=0.33333,
        use_cobaya=False,
        n_drag=0):
        """
        posterior should return a PipelineResults object
        """
        #Set up basic variables
        self.posterior = posterior
        self.p = np.array(start)
        self.ndim = len(self.p)
        self.quiet=quiet
        #Run the pipeline for the first time, on the 
        #starting point
        self.Lp = self.posterior(self.p)

        #Proposal
        self.covariance = covariance
        cholesky = np.linalg.cholesky(covariance)
        self.scaling = scaling
        self.exponential_probability = exponential_probability
        self.use_cobaya = use_cobaya
        self.n_drag = n_drag
        self.fast_indices = None
        self.slow_indices = None
        self.oversampling = None
        self.fast_slow_is_ready = False
        self.rng = np.random.default_rng()


        if self.n_drag > 0 and not self.use_cobaya:
            raise ValueError('You must set use_cobaya=T to have n_drag>0')

        if self.use_cobaya:
            from .proposal import cobaya_proposal
            # Initial proposer, without block - replace below
            self.proposal = cobaya_proposal.CobayaProposalWrapper(
                parameter_blocks=[np.arange(self.ndim)],
                oversampling_factors=[1],
                proposal_scale=self.scaling,
                random_state=self.rng,
                )
            self.proposal.set_covariance(covariance)
        else:
            self.proposal = Proposal(cholesky, scaling=scaling, exponential_probability=exponential_probability)

        #For adaptive sampling
        self.last_covariance_estimate = covariance.copy()       
        self.covariance_estimate = covariance.copy()
        self.chain = []
        self.n_cov_fail = 0

        #self.covariance_estimate.copy()
        self.mean_estimate = start.copy()
        self.tuning_frequency = tuning_frequency
        self.original_tuning_frequency = tuning_frequency
        self.tuning_grace = tuning_grace
        self.tuning_end = tuning_end

        #Set up instance variables storing samples, etc.
        self.samples = []
        self.iterations = 0
        self.accepted = 0
        self.accepted_since_tuning = 0
        self.iterations_since_tuning = 0


    def sample(self, n):
        samples = []
        blobs = []

        sample_method = (
            self._sample_dragging 
            if (self.fast_slow_is_ready and self.n_drag) 
            else self._sample_metropolis)
        #Take n sample mcmc steps

        samples = []
        for i in range(n):
            # update counts
            self.iterations += 1
            self.iterations_since_tuning += 1

            # generate a new sample
            s = sample_method()
            samples.append(s)
            # hack - should unify these
            self.chain.append(s.vector)

            if self.should_tune_now():
                self.tune()

        return samples

    def _sample_metropolis(self):
        # proposal point and its likelihood
        q = self.proposal.propose(self.p)
        #assume two proposal subsets for now
        Lq = self.posterior(q)
        if not self.quiet:
            print("  ".join(str(x) for x in q))
        #acceptance test
        delta = Lq.post - self.Lp.post
        if  accept(Lq.post, self.Lp.post):
            #update if accepted
            self.Lp = Lq
            self.p = q
            self.accepted += 1
            self.accepted_since_tuning += 1
            if not self.quiet:
                print("[Accept delta={:.3g}]\n".format(delta))
        elif not self.quiet:
            print("[Reject delta={:.3g}]\n".format(delta))

        return self.Lp


    def _sample_dragging(self):
        # get params with same fast params but different slow ones
        if not self.quiet:
            print("starting drag")
            print("Current post = ", self.Lp.post)
        start = self.p
        end = self.proposal.propose_slow(start)


        # posteriors and derived parameters etc.
        r_start = copy(self.Lp)
        r_end = self.posterior(end)
        if not self.quiet:
            print("slow proposal post = ", r_end.post)

        if not np.isfinite(r_end.post):
            if not self.quiet:
                print("[Reject: nan/-inf posterior]\n")
            return self.Lp


        # coordinates of current start and end
        p1 = copy(start)
        p2 = copy(end)

        # results for current start and end
        r1 = copy(r_start)
        r2 = copy(r_end)

        start_post = r1.post
        end_post = r2.post

        drag_accepts = 0

        for i in range(self.n_drag):
            delta_fast = self.proposal.propose_fast(p1) - p1
            if not self.quiet:
                print("delta fast", delta_fast)
            q1 = p1 + delta_fast
            q2 = p2 + delta_fast

            s1 = self.posterior(q1)

            if np.isfinite(s1.post):
                s2 = self.posterior(q2)
            else:
                s2 = Bad()


            f = (1+i) /(1+self.n_drag)

            P1 = (1-f)*r1.post + f*r2.post
            Q1 = (1-f)*s1.post + f*s2.post

            accept_drag = accept(Q1, P1) and np.isfinite(s1.post) and np.isfinite(s2.post)

            if accept_drag:
                p1 = q1
                p2 = q2
                r1 = s1
                r2 = s2
                drag_accepts += 1
                if not self.quiet:
                    print("[Accept drag step delta={:.3g}]\n".format(Q1 - P1))
            elif not self.quiet:
                print("[Reject drag step delta={:.3g}]\n".format(Q1 - P1))

            start_post += r1.post
            end_post += r2.post

        if not self.quiet:
            print("[Accepted {}/{} drag steps]".format(drag_accepts,self.n_drag))

        start_post /= self.n_drag
        end_post /= self.n_drag
        accept_overall = accept(end_post, start_post)

        if not self.quiet:
            print("Done drag")
        if accept_overall:
            self.p = p2
            self.Lp = r_end
            self.accepted += 1
            self.accepted_since_tuning += 1
            if not self.quiet:
                print("[Accept delta={:.3g}]\n".format(end_post - start_post))
            return r2
        else:
            if not self.quiet:
                print("[Reject delta={:.3g}]\n".format(end_post - start_post))
            return self.Lp



    def should_tune_now(self):
        return (    
            self.tuning_frequency>0
        and self.iterations>self.tuning_grace 
        and self.iterations%self.tuning_frequency==0
        and self.iterations<self.tuning_end
        )


    def update_covariance_estimate(self):
        n = self.iterations
        self.mean_estimate = np.mean(self.chain, axis=0)
        C = np.cov(np.transpose(self.chain))
        if is_positive_definite(C):
            self.covariance_estimate = C
        else:
            print("Cov estimate not SPD.  If this keeps happening, be concerned.")
            # chain_outfile = 'joe_dump_chain_{}.txt'.format(self.n_cov_fail)
            # cov_outfile = 'joe_dump_cov_{}.txt'.format(self.n_cov_fail)
            # self.n_cov_fail += 1
            # np.savetxt(chain_outfile, self.chain)
            # np.savetxt(cov_outfile, np.transpose(C))
            # print("TEMPORARY (JOE - REMOVE LATER) - dumping to file")


    def set_fast_slow(self, fast_indices, slow_indices, oversampling):
        if self.n_drag:
            oversampling = 1
            print("Overriding oversampling parameter -> 1 since using dragging")
        self.fast_indices = fast_indices
        self.slow_indices = slow_indices
        self.oversampling = oversampling
        self.fast_slow_is_ready = True

        if self.use_cobaya:
            from .proposal import cobaya_proposal
            self.proposal = cobaya_proposal.CobayaProposalWrapper(
                parameter_blocks=[self.slow_indices, self.fast_indices], 
                oversampling_factors=[1, oversampling],
                i_last_slow_block=0,
                proposal_scale=self.scaling,
                random_state=self.rng,
                )
            self.proposal.set_covariance(self.covariance)
        else:
            self.proposal = FastSlowProposal(self.covariance, fast_indices, slow_indices, oversampling, scaling=self.scaling, exponential_probability=self.exponential_probability)
        self.tuning_frequency = self.original_tuning_frequency * oversampling

    def tune(self):
        self.update_covariance_estimate()

        f = (self.covariance_estimate.diagonal()**0.5-self.last_covariance_estimate.diagonal()**0.5)/self.last_covariance_estimate.diagonal()**0.5
        i = abs(f).argmax()
        print("Largest parameter sigma fractional change = {:.1f}% for param {}".format(100*f[i], i))
        self.last_covariance_estimate = self.covariance_estimate.copy()

        print("Accepted since last tuning: {}%".format((100.*self.accepted_since_tuning)/self.iterations_since_tuning))
        self.accepted_since_tuning = 0
        self.iterations_since_tuning = 0


        if self.use_cobaya:
            print("Tuning cobaya proposal.")
            self.proposal.set_covariance(self.covariance_estimate)
        elif isinstance(self.proposal, FastSlowProposal):
            print("Tuning fast/slow sampler proposal.")
            self.proposal = FastSlowProposal(self.covariance_estimate, 
                self.fast_indices, self.slow_indices, self.oversampling,scaling=self.scaling, exponential_probability=self.exponential_probability)
        elif isinstance(self.proposal, Proposal):
            print("Tuning standard sampler proposal.")
            cholesky = np.linalg.cholesky(self.covariance_estimate)
            self.proposal = Proposal(cholesky, scaling=self.scaling, exponential_probability=self.exponential_probability)
        else:
            #unknown proposal type
            pass


def accept(post1, post0):
    return (post1 > post0) or (post1-post0 > np.log(np.random.uniform(0,1)))

def is_positive_definite(M):
    return np.all(np.linalg.eigvals(M) > 0)



import numpy as np

def setup(options):
    return {}

def execute(block, config):
    p1 = block['parameters', 'p1']
    p2 = block['parameters', 'p2']
    like = -(p1**2 + p2**2)/2.
    block['parameters', 'p3'] = p1 + p2
    block['likelihoods', 'test_like'] = like
    block['data_vector', 'test_theory'] = np.array([p1,p2])
    block['data_vector', 'test_inverse_covariance'] = np.array([[1., 0.,],[0., 1.,]])
    return 0
